/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.interpreter.prettyprint;

import java.util.*;

/**
 * Führt eine automatische Formatierung von Skripten durch.
 *
 * @author beck et al. projects GmbH
 * @author Yvonne Sunke
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class PrettyPrinter {

	/** Enthaelt alle Zeichenketten, vor denen ein Zeilenumbruch eingefuegt werden soll. */
	private static HashSet<String> NEW_LINE_BEFORE = new HashSet<>();

	/** Enthaelt alle Zeichenketten, nach denen ein Zeilenumbruch eingefuegt werden soll. */
	private static HashSet<String> NEW_LINE_AFTER = new HashSet<>();

	/** Enthaelt alle Zeichenketten, die als Schluesselwort definiert sind. */
	private static HashSet<String> ALL_KEYWORDS = new HashSet<>();

	/**
	 * Enthaelt alle Zeichenketten, die eine Kommentarzeile einleiten und nach denen kein Zeilenumbruch erfolgt. Erst eine EscapeSequenz zeigt an, dass die Zeile
	 * zu Ende ist.
	 */
	private static HashSet<String> ALL_COMMENT = new HashSet<>();

	/** Enthaelt alle relevanten Escape Sequenzen. */
	private static HashSet<String> ALL_ESCAPES = new HashSet<>();

	/** Enthaelt alle Zeichenketten, die das Ende einer Zeile markieren. */
	private static HashSet<String> ALL_END_OF_LINE = new HashSet<>();

	/** Enthaelt alle Zeichenketten, die eine <b>nachfolgende</b> -> Einrueckung verursachen. */
	private static HashSet<String> TAB_PLUS = new HashSet<>();

	/** Enthaelt alle Zeichenketten, die eine <b>vorherige</b> <- Einrueckung verursachen. */
	private static HashSet<String> TAB_MINUS = new HashSet<>();

	/** Enthaelt alle Zeichenketten, die eine <b>vorherige</b> Einrueckung komplett aufheben. */
	private static HashSet<String> TAB_CLEAR = new HashSet<>();

	/** Leere Zeichenkette, die als Einrückung dient. */
	private static String EMPTY_TAB = "   ";

	/**
	 * Intitialisierung aller statischer HashSets.
	 */
	static {
		// Zeilenendemarkierungen. Oder besser Zeichen, die einen Zeilenumbruch bewirken.
		PrettyPrinter.ALL_END_OF_LINE.add("\n");
		PrettyPrinter.ALL_END_OF_LINE.add("\f");
		PrettyPrinter.ALL_END_OF_LINE.add("\r");

		// Escape Sequenzen. Muessen entweder als Ende-Marker eines Kommentars
		// interpretiert oder einfach gelöscht werden.
		PrettyPrinter.ALL_ESCAPES.add("\t");
		PrettyPrinter.ALL_ESCAPES.addAll(ALL_END_OF_LINE);

		// Alle Zeichenketten, die wie Kommentarzeilen behandelt werden. Zeilenumbruch passiert erst bei einem Objekt,
		// das in <code>allEndOfLine</code> enthalten ist.
		PrettyPrinter.ALL_COMMENT.add("#REGEL#");
		for(int i = 1; i <= 100; i++) { // Ich nehme mal an, dass niemand mehr als 100 #-Zeichen in eine Zeile schreibt.
			StringBuffer tmp = new StringBuffer();
			for(int j = 1; j <= i; j++) {
				tmp.append("#");
			}
			PrettyPrinter.ALL_COMMENT.add(tmp.toString());
		}

		// Zeilenumbrueche vor folgenden Worten.
		PrettyPrinter.NEW_LINE_BEFORE.add("#REGEL#"); // VRZ3 Entwicklerteam intern
		PrettyPrinter.NEW_LINE_BEFORE.add("filter");
		PrettyPrinter.NEW_LINE_BEFORE.add("spalten");
		PrettyPrinter.NEW_LINE_BEFORE.add("standards");
		PrettyPrinter.NEW_LINE_BEFORE.add("definitionen");
		PrettyPrinter.NEW_LINE_BEFORE.add("aggregationsfunktion");
		PrettyPrinter.NEW_LINE_BEFORE.add("alias");
		PrettyPrinter.NEW_LINE_BEFORE.add("anwendung");
		PrettyPrinter.NEW_LINE_BEFORE.add("attribut");
		PrettyPrinter.NEW_LINE_BEFORE.add("attributgruppe");
		PrettyPrinter.NEW_LINE_BEFORE.add("einschr\u00e4nkung");
		PrettyPrinter.NEW_LINE_BEFORE.add("ende mit");
		PrettyPrinter.NEW_LINE_BEFORE.add("filter");
		PrettyPrinter.NEW_LINE_BEFORE.add("mit");
		PrettyPrinter.NEW_LINE_BEFORE.add("nachfilter");
		PrettyPrinter.NEW_LINE_BEFORE.add("protokollart");
		PrettyPrinter.NEW_LINE_BEFORE.add("protokollzeitraum");
		PrettyPrinter.NEW_LINE_BEFORE.add("pseudoobjekt");
		PrettyPrinter.NEW_LINE_BEFORE.add("symbol");
		PrettyPrinter.NEW_LINE_BEFORE.add("tempAttribut");
		PrettyPrinter.NEW_LINE_BEFORE.add("verschmelzen");

		// Zeilenumbruch nach folgenden Worten.
		PrettyPrinter.NEW_LINE_AFTER.add("spalten");
		PrettyPrinter.NEW_LINE_AFTER.add("filter");
		PrettyPrinter.NEW_LINE_AFTER.add("definitionen");
		PrettyPrinter.NEW_LINE_AFTER.add("standards");

		// Alle Schluesselworte
		PrettyPrinter.ALL_KEYWORDS.add("aenderungsprotokoll");
		PrettyPrinter.ALL_KEYWORDS.add("\u00e4nderungsprotokoll");
		PrettyPrinter.ALL_KEYWORDS.add("zustandsprotokoll");
		PrettyPrinter.ALL_KEYWORDS.add("?"); // Sonderfall fuer Zeilenumbruch
		PrettyPrinter.ALL_KEYWORDS.add("abs");
		PrettyPrinter.ALL_KEYWORDS.add("aggregation");// Sonderfall fuer Zeilenumbruch
		PrettyPrinter.ALL_KEYWORDS.add("als");
		PrettyPrinter.ALL_KEYWORDS.add("div");
		PrettyPrinter.ALL_KEYWORDS.add("exp");
		PrettyPrinter.ALL_KEYWORDS.add("falsch");
		PrettyPrinter.ALL_KEYWORDS.add("gesamt");
		PrettyPrinter.ALL_KEYWORDS.add("intervall");
		PrettyPrinter.ALL_KEYWORDS.add("istFehler");
		PrettyPrinter.ALL_KEYWORDS.add("istZahl");
		PrettyPrinter.ALL_KEYWORDS.add("istZustand");
		PrettyPrinter.ALL_KEYWORDS.add("istLeer");
		PrettyPrinter.ALL_KEYWORDS.add("istText");
		PrettyPrinter.ALL_KEYWORDS.add("istWahrheitswert");
		PrettyPrinter.ALL_KEYWORDS.add("liste");
		PrettyPrinter.ALL_KEYWORDS.add("ln");
		PrettyPrinter.ALL_KEYWORDS.add("max");
		PrettyPrinter.ALL_KEYWORDS.add("modulo");
		PrettyPrinter.ALL_KEYWORDS.add("nicht");
		PrettyPrinter.ALL_KEYWORDS.add("objekt");
		PrettyPrinter.ALL_KEYWORDS.add("oder");
		PrettyPrinter.ALL_KEYWORDS.add("potenz");
		PrettyPrinter.ALL_KEYWORDS.add("spalte");
		PrettyPrinter.ALL_KEYWORDS.add("und");
		PrettyPrinter.ALL_KEYWORDS.add("wahr");
		PrettyPrinter.ALL_KEYWORDS.addAll(NEW_LINE_BEFORE);
		PrettyPrinter.ALL_KEYWORDS.addAll(NEW_LINE_AFTER);
		PrettyPrinter.ALL_KEYWORDS.addAll(ALL_COMMENT);

		// Alle Zeichenketten, die eine Einrückung nach rechts verursachen.
		PrettyPrinter.TAB_PLUS.add("spalten");
		PrettyPrinter.TAB_PLUS.add("filter");
		PrettyPrinter.TAB_PLUS.add("definitionen");
		PrettyPrinter.TAB_PLUS.add("standards");
		PrettyPrinter.TAB_PLUS.add("mit");

		// Alle Zeichenketten, die eine Einrückung nach links verursachen.
		PrettyPrinter.TAB_MINUS.add("ende mit");

		// Alle Zeichenketten, die eine Einrückung komplett aufheben.
		PrettyPrinter.TAB_CLEAR.add("spalten");
		PrettyPrinter.TAB_CLEAR.add("filter");
		PrettyPrinter.TAB_CLEAR.add("definitionen");
		PrettyPrinter.TAB_CLEAR.add("standards");
	}

	/**
	 * Formatiert den übergebenen Text wie in TAnf beschrieben.
	 *
	 * @param source Original Text.
	 *
	 * @return Frisch umformatierten Text.
	 */
	public static String prettyPrint(String source) {
		PrettyPrinter pp = new PrettyPrinter();
		return pp.doTheFormatting(source);
	}

	/**
	 * Erledigt die Neuformatierung.
	 *
	 * @param source Original Text
	 *
	 * @return Frisch umformatierten Text.
	 */
	private String doTheFormatting(String source) {
		int noOfTabs = 0;
		StringBuffer prettyResult = new StringBuffer();
		boolean endOfSource = false;

		while(!endOfSource) {
			String keyword = findKeyword(source);
			if(!"".equals(keyword)) // noch mindestens ein Keyword da
			{
				String text = findTextBefore(source, keyword);

				// TEXT VOR DEM KEYWORD , Evtl. Leerzeichen und dann den "trimmed" Text anfuegen.
				if(text.trim().length() > 0) {
					if(prettyResult.length() > 0) {
						String prettyEnde = prettyResult.substring(prettyResult.length() - 1, prettyResult.length());
						if("?".equals(prettyEnde)) {/* Nix tun */
						}
						else if(!"\n".equals(prettyEnde)) {
							prettyResult.append(" ");
						}
						else {
							for(int i = 1; i <= noOfTabs; i++) {
								prettyResult.append(PrettyPrinter.EMPTY_TAB);
							}
						}
					}
					prettyResult.append(text.trim());
				}

				// Evtl. ZEILENUMBRUCH VOR DEM KEYWORD
				if(prettyResult.length() > 0) {
					String prettyEnde = prettyResult.substring(prettyResult.length() - 1, prettyResult.length());
					if(PrettyPrinter.NEW_LINE_BEFORE.contains(keyword) && !"\n".equals(prettyEnde)) {
						prettyResult.append("\n");
					}
					else if(PrettyPrinter.ALL_COMMENT.contains(keyword) && !"\n".equals(prettyEnde)) {
						prettyResult.append("\n");

						// TODO das funktioniert noch nicht zuverlaessig genug.
						// // Kommentarzeilen extra behandeln. Manche brauchen einen Zeilenumbruch, manche nicht.
						// int idx = findHighestEOLIdx(text); //Das letzte EOL Zeichen vor dem keyword
						// if(idx > -1) {
						// String tmp = text.substring(idx).trim(); // trim
						// if("".equals(tmp)) // wenn zwischen EOL und Kommentar nichts ist
						// prettyResult.append("\n"); // dann Zeilenumbruch.
						// }
						// else if(!"\n".equals(prettyEnde))
						// prettyResult.append(EMPTY_TAB).append(EMPTY_TAB).append(EMPTY_TAB);
					}
				}

					if("?".equals(keyword)) {
					final int keyWordIndex = source.indexOf(keyword);
					final String space;
					if(keyWordIndex > 0){
						final String substring = source.substring(keyWordIndex - 1, keyWordIndex);
						space = keyWordIndex == 0 ? "" : substring.equals(" ") || ALL_ESCAPES.contains(substring) ? " " : "";
					}
					else{
						space = "";
					}

					source = source.substring(keyWordIndex); // alles bis vor "?" abschneiden
					// Sonderfall: ?123=<Wort> , ?213 = <Wort> , ?213 =<Wort> , ?123= <Wort>
					// Achtung: ?123 <Wort> ist KEIN! Sonderfall
					int idxSpace = findLowestEscapeOrEmptyIdx(source);
					int idxEqual = source.indexOf("=");
					int idxEnd = 1; // Ende Idx faengt ein Zeichen hinter dem ? an.
					if(idxSpace > 1) {
						idxEnd = idxSpace; // Leerzeichen kommt nach idxEnd, aber noch vor dem =.
						if(idxEqual > 1 && idxEqual < idxSpace) idxEnd = idxEqual; // = kommt nach idxEnd, aber noch vor dem nächsten Leerzeichen.
					}
					else if(idxEqual > 1) idxEnd = idxEqual;
					String number = source.substring(1, idxEnd);
					try {
						Integer.parseInt(number);
						source = source.substring(idxEnd); // '?13223' abschneiden
						source = source.trim();
						String next = source.substring(0, 1);
						if("=".equals(next)) {
							String prettyEnde = prettyResult.substring(prettyResult.length() - 1, prettyResult.length());
							if(!"\n".equals(prettyEnde)) prettyResult.append("\n");
							for(int i = 1; i <= noOfTabs; i++) {
								prettyResult.append(PrettyPrinter.EMPTY_TAB);
							}
							prettyResult.append(keyword).append(number).append(" = ");
							// alles super. Wort suchen
							source = source.substring(1); // ein zeichen abschneiden
							source = source.trim(); // leerzeichen entfernen
							int idx = findLowestEscapeOrEmptyIdx(source);
							next = source.substring(0, idx).trim(); // zeichenkette bis zum nächsten leerzeichen suchen
							// und vorsichtshalber trim() (wg. \t)
							prettyResult.append(next);
							if(!findKeyword(source).startsWith("#")) prettyResult.append("\n");
							// Vorsicht! Bei indexOf-Aktionen immer nach LowerCase konvertieren!
							// SOURCE KUERZEN! Um den Text, EscapeSequenzen und das Keyword
							source = source.substring(next.length()).trim();
						}
						else {
							// war wohl doch kein keyword. einfach weitermachen wie bisher.
							prettyResult.append(space).append(keyword).append(number);
							// source hier nicht mehr kürzen. ist schon vor dem if/else passiert!
						}
					}
					catch(Exception e) {
						prettyResult.append(space).append(keyword).append(number);
						// Vorsicht! Bei indexOf-Aktionen immer nach LowerCase konvertieren!
						// SOURCE KUERZEN! Um den Text, EscapeSequenzen und das Keyword
						source = source.substring(idxEnd).trim();
					}
				}
				else {
					// Checken ob Einrueckung VOR dem keyword reduziert werden muss.
					if(PrettyPrinter.TAB_CLEAR.contains(keyword)) {
						noOfTabs = 0;
					}
					else if(PrettyPrinter.TAB_MINUS.contains(keyword)) noOfTabs--;

					if(prettyResult.length() > 0) {
						String prettyEnde = prettyResult.substring(prettyResult.length() - 1, prettyResult.length());
						if("\n".equals(prettyEnde)) {
							for(int i = 1; i <= noOfTabs; i++)
							// Einrueckung andrucken
							{
								prettyResult.append(PrettyPrinter.EMPTY_TAB);
							}
						}
						else {
							prettyResult.append(" ");
						}
					}
					// Keyword einfuegen
					prettyResult.append(keyword);
					// Checken ob nachfolgende Einrueckung erweitert werden muss.
					if(PrettyPrinter.TAB_PLUS.contains(keyword)) noOfTabs++;

					// SOURCE KUERZEN! Um den Text, EscapeSequenzen und das Keyword
					source = source.substring(source.indexOf(keyword) + keyword.length());

					// Evtl. ZEILENUMBRUCH NACH DEM KEYWORD.
					if(PrettyPrinter.NEW_LINE_AFTER.contains(keyword)) {
						prettyResult.append(":\n");
						source = source.substring(source.indexOf(":") + 1);
					}
				}

				// Einige Zeilen duerfen nicht umgebrochen werden. Erst, wenn ein wirkliches Zeilenumbruch-Zeichen
				// gefunden wird.
				// Kommentar und #REGEL# Zeilen
				if(PrettyPrinter.ALL_COMMENT.contains(keyword)) {
					int idx = findLowestEOLIdx(source);
					String txt = source.substring(0, idx);
					prettyResult.append(" ").append(txt.trim()).append("\n");
					source = source.substring(idx).trim();
				}
			}
			else
			// kein Keyword mehr vorhanden
			{
				endOfSource = true;
				String text = source;
				// TEXT NACH DEM allerletzten KEYWORD , Evtl. Leerzeichen und dann den "trimmed" Text anfuegen.
				if(text.trim().length() > 0) {
					if(prettyResult.length() > 0) {
						String prettyEnde = prettyResult.substring(prettyResult.length() - 1, prettyResult.length());
						if(!"\n".equals(prettyEnde)) {
							prettyResult.append(" ");
						}
						else {
							for(int i = 1; i <= noOfTabs; i++) {
								prettyResult.append(PrettyPrinter.EMPTY_TAB);
							}
						}
					}
					prettyResult.append(text.trim());
				}
			}
		}
		return prettyResult.toString();
	}

	private int findLowestEscapeOrEmptyIdx(String source) {
		int lowestIdx = source.length();
		for(String key : PrettyPrinter.ALL_ESCAPES) {
			int idx = source.indexOf(key);
			if(idx > -1 && idx < lowestIdx) lowestIdx = idx;
		}
		int emptyidx = source.indexOf(" ");
		if(emptyidx > -1 && emptyidx < lowestIdx) lowestIdx = emptyidx;
		return lowestIdx;
	}

	private int findLowestEOLIdx(String source) {
		int lowestIdx = source.length();
		for(String key : PrettyPrinter.ALL_END_OF_LINE) {
			int idx = source.indexOf(key);
			if(idx > -1 && idx < lowestIdx) lowestIdx = idx;
		}
		return lowestIdx;
	}

	// private int findHighestEOLIdx(String source) {
	// int highestIdx = -1;
	// for (String key : PrettyPrinter.ALL_END_OF_LINE) {
	// int idx = source.lastIndexOf(key);
	// if(idx > highestIdx)
	// highestIdx = idx;
	// }
	// return highestIdx;
	// }

	/**
	 * Sucht im uebergebenen source String nach dem Text, der sich vor dem uebergebenen keyword befindet.
	 *
	 * @param source  Die Quelle.
	 * @param keyword Das Schluesselwort.
	 *
	 * @return Den Text innerhalb der Quelle vor dem Schluesselwort.
	 */
	private String findTextBefore(String source, String keyword) {
		return source.substring(0, source.indexOf(keyword));
	}

	/**
	 * Sucht im uebergebenen source String nach dem naechsten Schluesselwort.
	 *
	 * @param source Die Quelle.
	 *
	 * @return Das naechste Schluesselwort innerhalb der Quelle.
	 */
	private String findKeyword(String source) {
		boolean keywordFound = false;
		String keyword = "";

		while(!keywordFound) {
			int lowestIdx = source.length();
			keyword = "";
			for(String key : ALL_KEYWORDS) {
				int idx = source.indexOf(key);
				if(idx > -1 && idx < lowestIdx) {
					lowestIdx = idx;
					keyword = key;
				}
				else if(idx > -1 && idx == lowestIdx) {
					if(key.length() > keyword.length()) keyword = key; // ueberschreiben falls das gefundene Keyword mehr Zeichen hat.
				}
			}
			// Hier muss noch gecheckt werden, dass das gefundene wort tatsächlich ein keywort ist
			// und nicht teil eines zusammengesetzten bezeichners. zB alias und mein_alias oder mit und mitarbeiter
			if(keyword.startsWith("#") || "?".equals(keyword)) {
				break;
			}
			boolean zeichenVorOK = false;
			if(lowestIdx > 0) {
				String zeichenVor = source.substring(lowestIdx - 1, lowestIdx);
				if(" ".equals(zeichenVor)) {
					zeichenVorOK = true;
				}
				else {
					for(String zeichen : PrettyPrinter.ALL_ESCAPES) {
						if(zeichen.equals(zeichenVor)) zeichenVorOK = true;
					}
				}
			}
			else {
				zeichenVorOK = true; // Wir sind gaaanz am anfang
			}

			boolean zeichenNachOK = false;
			if(lowestIdx + keyword.length() + 1 <= source.length()) {
				String zeichenNach = source.substring(lowestIdx + keyword.length(), lowestIdx + keyword.length() + 1);
				if(" ".equals(zeichenNach) || ":".equals(zeichenNach) || "(".equals(zeichenNach)) {
					zeichenNachOK = true;
				}
				else {
					for(String zeichen : PrettyPrinter.ALL_ESCAPES) {
						if(zeichen.equals(zeichenNach)) zeichenNachOK = true;
					}
				}
			}
			else {
				zeichenNachOK = true;
			}

			if(zeichenVorOK && zeichenNachOK) {
				keywordFound = true;
			}
			else {
				source = source.substring(lowestIdx + keyword.length());
			}
		}
		return keyword;
	}
}
