/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.interpreter.semantics;

import antlr.collections.AST;
import de.bsvrz.pua.prot.exceptions.CriticalParserException;
import de.bsvrz.pua.prot.exceptions.SemanticErrorException;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.pua.prot.util.attributes.AttributeDescription;
import de.bsvrz.pua.prot.util.attributes.RealAttributeDescription;
import de.bsvrz.pua.prot.util.attributes.TempAttributeDescription;

import java.util.*;

/**
 * Überprüf die Semantik von in Ausdrücken eingetragenen Attributen.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
@SuppressWarnings("deprecation")
public class ExpressionSemantics implements SemanticChecks {

	private SymbolTable _symbolTable;

	private AttributeSemantics _attributeSemantics;

	private RealAttributeSemantics _realAttributeSemantics;

	private Map<TempAttributeDescription, AST> _tempAttributesUsedInFilterExpression = new HashMap<>();

	/**
	 * @param semantics Semantics
	 * @param attributeSemantics AttributeSemantics
	 * @param realAttributeSemantics RealAttributeSemantics
	 */
	public ExpressionSemantics(final Semantics semantics, final AttributeSemantics attributeSemantics, final RealAttributeSemantics realAttributeSemantics) {
		this._symbolTable = semantics.symbolTable;
		this._attributeSemantics = attributeSemantics;
		this._realAttributeSemantics = realAttributeSemantics;
	}

	/**
	 * Prüft ein Attribut, das in einem Ausdruck vorkommt und fügt es zu den Informationen für die Datenaufbereitung hinzu.
	 *
	 * @param ast             Knoten (wird nur für Zeilenangabe in der Fehlermeldung benötigt)
	 * @param parts           Die Qualifizierung des Attributs
	 * @param isTempAttribute false: Attribut in einem Filter-Ausdruck verwendet. true: Attribut wird bei der Definition eines temporären Attributs verwendet
	 * @param padData         Falls das Auffüllen von Daten unterbleiben soll, muss <code>padData</code> ungleich <code>null</code> sein.
	 *
	 * @return AttributeDescription Attributbeschreibung eines realen oder temporären Attributs
	 *
	 * @throws CriticalParserException Falls das Attribut nicht identifiziert werden konnte
	 */
	public AttributeDescription[] createExpressionAttribute(final AST ast, final List<String> parts, final boolean isTempAttribute, final AST padData)
			throws CriticalParserException {
		try {
			//überprüfen, ob es sich um ein Reales oder temporäres Attribut handelt:
			if(parts.size() == 1) {
				final String attribute = parts.get(0);
				if(_symbolTable.isTempAttributeDefined(attribute)) {
					//temporäres Attribut
					//existiert schon -> aus der Symboltabelle holen
					final TempAttributeDescription tempAtt = _symbolTable.getTempAttributeDefinition(attribute);

					if(tempAtt == null) {
						throw new SemanticErrorException(InterpreterErrorMessage.UNDEFINED_TEMPATT + attribute, ast);
					}
					else {
						if(!isTempAttribute) {
							_tempAttributesUsedInFilterExpression.put(tempAtt, ast);
						}
						return new AttributeDescription[]{tempAtt};
					}
				}
			}

			//keine Attributliste (oder kein Element angegeben)
			final RealAttributeDescription[] ratt = _attributeSemantics.storeAttributeInformation(
					_realAttributeSemantics.createRealAttributeDescription(
							parts, (padData == null)
					)
			);

			//prüfen, ob atomar:
			if(!ratt[0].isAtomic()) throw new SemanticErrorException(InterpreterErrorMessage.ATOM_EXPECTED + parts);

			if(isTempAttribute) //Attribut wird in einem temporären Attribut verwendet
			{
				_attributeSemantics.incAttributeCount();
			}

			return ratt;
		}
		catch(CriticalParserException criticalException) {
			criticalException.line = ast.getLine();
			throw criticalException;
		}
	}

	/**
	 * @see de.bsvrz.pua.prot.interpreter.semantics.SemanticChecks#cleanUp(de.bsvrz.pua.prot.util.ProcessingParameter,boolean)
	 */
	public void cleanUp(final ProcessingParameter pp, final boolean processScript) throws SemanticErrorException {
	}
}
