/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.processing.util;

import de.bsvrz.pua.prot.client.dataobject.PuADataState;
import de.bsvrz.pua.prot.util.ExpressionResult;
import de.bsvrz.pua.prot.util.ResultValue;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.util.ArrayList;
import java.util.List;

/**
 * Baustein für den Ausgangsdatensatz. Enthält den Wert eines Attributs und die für weitere Berechnungen wichtigen Werte. Wird im Ausgangsdatensatz ({@link
 * de.bsvrz.pua.prot.processing.util.IntermediateDataSet}) verwendet.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class BaseDataSet {

	/** Datenzeitstempel oder Datenindex wurde nicht gesetzt. */
	public static final long INVALID_VALUE = -1;

	private final ResultValue _value;

	private final DataInformation _dataInformation;

	private final long _dataTimeStamp;

	private final long _archiveTimeStamp;

	private final long _startTimeStamp;

	private final long _endTimeStamp;

	private final long _dataIndex;

	private final boolean _paddedData;
	
	private final PuADataState _dataState;

	/**
	 * Erzeugt einen Datensatz.
	 *  @param value            Wert des Elements
	 * @param dataState        Datenzustand
	 * @param dataInformation  Markierungsinformation
	 * @param dataTimeStamp    Datenzeitstempel
	 * @param archiveTimeStamp Archivzeitstempel. Ist bei Online-Daten und temporären Attributen  {@link #INVALID_VALUE}.
	 * @param dataIndex        Datenindex. Ist bei temporären Attributen {@link #INVALID_VALUE}.
	 * @param paddedData       Daten wurden aufgefüllt. Hat nur Informationswert für die {@link #toString()}-Methode - keinerlei Funktion.
	 * @param endTimeStamp endTimeStamp
	 * @param startTimeStamp startTimeStamp
	 */
	public BaseDataSet(ResultValue value, final PuADataState dataState, DataInformation dataInformation, long dataTimeStamp, long archiveTimeStamp, long startTimeStamp, long endTimeStamp, long dataIndex, boolean paddedData) {
		if(dataState == null) throw new IllegalArgumentException("dataState ist null");
		this._value = value;
		this._dataState = dataState;
		this._dataInformation = dataInformation;
		this._dataTimeStamp = dataTimeStamp;
		this._archiveTimeStamp = archiveTimeStamp;
		this._startTimeStamp = startTimeStamp;
		this._endTimeStamp = endTimeStamp;
		this._dataIndex = dataIndex;
		this._paddedData = paddedData;
	}

	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		StringBuffer sb = new StringBuffer();
		if(_paddedData == false) {
			sb.append("[");
			Util.msToDate(sb, _dataTimeStamp);
			sb.append(" (" + _dataTimeStamp + "), #" + _dataIndex + ": ");
			if(_value != null) {
				sb.append(_value.toString());
			}
			else {
				sb.append("no data");
			}
			sb.append(" ]");
		}
		else {
			sb.append("[aufgefüllte Daten: #" + _dataTimeStamp + ", " + _dataIndex + "]");
		}
		if(_dataInformation != null) sb.append(" -> M: " + _dataInformation);
		return sb.toString();
	}

	/**
	 * Vergleicht Ausgangsdatensätze. Beim Vergleich werden nur die Werte der Daten und die Markierung verglichen. Datenzeitstempel und Datenindex werden dabei
	 * nicht berücksichtigt.
	 *
	 * @param first  Ausgangsdatensatz
	 * @param second Ausgangsdatensatz
	 *
	 * @return	<code>true</code> falls die Datensätze gleich sind.<code>false</code> sonst.
	 */
	public static boolean compareBaseData(List<BaseDataSet> first, List<BaseDataSet> second) {
		if(first == null && second == null)		//beide null
		{
			return true;
		}
		else if(first == null || second == null)	//eins von beiden ist nicht null
		{
			return false;
		}
		else if(first.size() == second.size())							//beide != null
		{
			boolean result = true;
			for(int i = 0; i < first.size() && result; i++) {
				BaseDataSet currentData = first.get(i);
				BaseDataSet lastData = second.get(i);
				if(currentData == null && lastData == null){
					// Beide null, also gleich, also einfach weitermachen
				}
				else if(currentData == null || lastData == null){
					return false; // Nur einer Null, also ungleich.
				}
				else if(currentData.getValue() == null && lastData.getValue() == null){
					// Beide null, also gleich, also einfach weitermachen
				}
				else if(currentData.getValue() == null || lastData.getValue() == null){
					return false;
				}
				else{
					result = currentData.getValue().equals(lastData.getValue())
							 && currentData.getDataInformation().cmpInfo(lastData.getDataInformation());
				}
			}
			return result;
		}
		else {
			return false;
		}
	}


	public static List<BaseDataSet> createDiffDataSet(final List<BaseDataSet> baseData, final List<BaseDataSet> lastBaseData) {
		if(baseData == null) throw new IllegalArgumentException("baseData ist null");
		if(lastBaseData == null) return baseData;

		if(lastBaseData.size() != baseData.size()) throw new IllegalStateException("List size");

		ArrayList<BaseDataSet> result = new ArrayList<>(baseData);
		for(int i = 0; i < baseData.size(); i++) {
			final BaseDataSet currentData = baseData.get(i);
			final BaseDataSet lastData = lastBaseData.get(i);
			boolean dataEqual = false;
			if(currentData != null && currentData.getValue() != null
					&& lastData != null && lastData.getValue() != null){
				dataEqual = currentData.getValue().equals(lastData.getValue());
			}
			if(dataEqual){
				BaseDataSet noChangeDataSet = new BaseDataSet(
						new ResultValue(ExpressionResult.noChange()),
						PuADataState.NO_CHANGE,
						currentData.getDataInformation(),
						currentData.getDataTimeStamp(),
						currentData.getArchiveTimeStamp(),
						currentData.getStartTimeStamp(),
						currentData.getEndTimeStamp(),
						currentData.getDataIndex(),
						currentData.isPaddedData()
				);
				result.set(i, noChangeDataSet);
			}
		}
		return result;
	}

	/** Zeigt an ob die Daten aufgefüllt wurden. */ /**
	 * Handelt es sich um aufgefüllte Daten?
	 *
	 * @return	<code>true</code> falls es sich um aufgefüllte Daten handelt.
	 */
	public boolean isPaddedData() {
		return _paddedData;
	}

	/** Wert des Elements */ /**
	 * Liefert den Wert des Attributs zurück.
	 *
	 * @return Wert des Attributs.
	 */
	public ResultValue getValue() {
		return _value;
	}

	/** Archivzeitstempel. Ist bei temporären Attributen {@link #INVALID_VALUE}. */ /**
	 * Archivzeitstempel. Ist bei temporären Attributen {@link #INVALID_VALUE}.
	 *
	 * @return Archivzeitstempel.
	 */
	public long getArchiveTimeStamp() {
		return _archiveTimeStamp;
	}

	/** Datenindex. Ist bei temporären Attributen {@link #INVALID_VALUE}. */ /**
	 * Datenindex. Ist bei temporären Attributen {@link #INVALID_VALUE}.
	 *
	 * @return Datenindex.
	 */
	public long getDataIndex() {
		return _dataIndex;
	}

	/** Markierungsinformation */ /**
	 * Markierungsinformation.
	 *
	 * @return Markierungsinformation.
	 */
	public DataInformation getDataInformation() {
		return _dataInformation;
	}

	/** Datenzeitstempel */ /**
	 * Datenzeitstempel.
	 *
	 * @return Datenzeitstempel.
	 */
	public long getDataTimeStamp() {
		return _dataTimeStamp;
	}

	public long getStartTimeStamp() {
		return _startTimeStamp;
	}

	public long getEndTimeStamp() {
		return _endTimeStamp;
	}

	public PuADataState getDataState() {
		return _dataState;
	}

}
