/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.util;

import java.io.Serializable;
import java.text.NumberFormat;
import java.util.Locale;

/**
 * Enthält Werte die von PuA verarbeitet und an den Client gesendet werden. Dabei handelt es sich entweder um Werte direkt vom Datenempfänger oder um das
 * Ergebnis der Auswertung eines {@link ExpressionTree}
 * <p>
 * Folgende Werte sind möglich:
 * <ul>
 * <li>Boolean (Ja/Nein)</li>
 * <li>Long (Unskalierte Integer-Werte)</li>
 * <li>Double (Skalierte Integer-Werte)</li>
 * <li>String (Text)</li>
 * <li>Status (Spezielle Statuswerte ausßerhalb des Wertebereichs, vgl. {@link de.bsvrz.dav.daf.main.config.IntegerValueState})</li>
 * <li>Error (Fehlermeldungen und -werte)</li>
 * <li>None (Kein Wert)</li>
 * </ul>
 * Es soll eine automatische Konvertierung der Werte erfolgen, so dass auch andere Methoden einen Wert zurückliefern. Die möglichen Konvertierungen werden in
 * den jeweiligen get-Methoden beschrieben.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public final class ExpressionResult implements Serializable {

	private static final long serialVersionUID = -6661919871655091243L;

	private static final ThreadLocal<NumberFormat> _numberFormat = new ThreadLocal<NumberFormat>(){
		// NumberFormat ThreadLocal da nicht Threadsafe
		@Override
		protected NumberFormat initialValue() {
			NumberFormat numberFormat = NumberFormat.getInstance(Locale.GERMANY);
			numberFormat.setGroupingUsed(false);
			return numberFormat;
		}
	};

	/** Typ des Ergebniswerts */
	private ResultType _type = ResultType.NONE;

	private boolean _bool = false;

	private long _long = 0;

	private double _double = 0.0;

	private String _string = null;

	/** Skalierungsfaktor. Wird nur noch PuA-Seitig benutzt und ist dort verzichtbar. */
	@Deprecated
	private double _conversionFactor = 1.0;

	/** True wenn es sich um einen Statuswert handelt. Wird nur noch PuA-Seitig zur ausgabe von toString() verwendet. */
	@Deprecated
	private boolean _isState = false;

	/** Erzeugt ein ExpressionResult mit dem Typ {@link ResultType#NONE}. */
	public ExpressionResult() {
	}

	/**
	 * Erzeugt ein ExpressionResult mit dem Typ {@link ResultType#STRING}.
	 *
	 * @param message Text, der abgelegt wird.
	 */
	public ExpressionResult(final String message) {
		this.set(message);
	}

	/**
	 * Erzeugt ein ExpressionResult mit dem Typ {@link ResultType#LONG}, also einer ganzen Zahl.
	 *
	 * @param l Zahl, die abgelegt wird.
	 *          
	 * @deprecated Funktion macht keine Unterscheidung bei skalierten Werten 
	 */
	@Deprecated
	public ExpressionResult(final long l) {
		setUnscaledValue(l, 1.0);
	}

	/**
	 * Erzeugt ein ExpressionResult mit dem Typ {@link ResultType#LONG}, also einer ganzen Zahl.
	 *
	 * @param l                Zahl, die abgelegt wird.
	 * @param conversionFactor Skalierungsfaktor für den Double-Wert
	 */
	public ExpressionResult(final long l, final double conversionFactor) {
		setUnscaledValue(l, conversionFactor);
	}

	/**
	 * Erzeugt ein ExpressionResult mit einem Fließkomma-Wert.
	 *
	 * @param d Zahl, die abgelegt wird.
	 */
	public ExpressionResult(final double d) {
		set(d);
	}

	/**
	 * Erzeugt ein ExpressionResult mit einem Bool-Wert.
	 *
	 * @param b Wert
	 */
	public ExpressionResult(final boolean b) {
		set(b);
	}

	/**
	 * Setzt den Wert dieses ExpressionResults auf den angegebenen Wahrheitswert.
	 *
	 * @param b Wert
	 */
	public void set(final boolean b) {
		setAll(ResultType.BOOL, b ?1.0 : 0.0, b ? 1 : 0, b ? "true" : "false", b, false);
	}

	/**
	 * Setzt den Wert dieses ExpressionResults auf den angegebenen Long-Wert.
	 *
	 * @param l Wert
	 */
	public void setUnscaledValue(final long l) {
		setAll(ResultType.LONG, (double)l, l, Long.toString(l), false, false);
	}

	/**
	 * Setzt den Wert dieses ExpressionResults auf den angegebenen Long-Wert.
	 *
	 * @param l                Wert
	 * @param conversionFactor Skalierungsfaktor für den Double-Wert
	 */
	public void setUnscaledValue(final long l, final double conversionFactor) {
		_conversionFactor = conversionFactor;
		if(_conversionFactor == 1) {
			setAll(ResultType.LONG, (double)l, l, formatLong(l), false, false);
		}
		else {
			setAll(ResultType.LONG, (double)l * conversionFactor, l, formatDouble(l * conversionFactor), false, false);
		}
	}

	/**
	 * Setzt den Wert dieses ExpressionResults auf den angegebenen Double-Wert.
	 *
	 * @param d Wert
	 */
	public void set(final double d) {
		setAll(ResultType.DOUBLE, d, Math.round(d), formatDouble(d), false, false);
	}

	/**
	 * Setzt den Wert dieses ExpressionResults auf den angegebenen String-Wert.
	 *
	 * @param s Wert
	 */
	public void set(final String s) {
		setAll(ResultType.STRING, 0.0, (long)0, s, false, false);
	}

	/**
	 * Übernimmt den Wert des übergebenen ExpressionResults. Die Werte werden kopiert. Es erfolgt keine Referenzierung.
	 *
	 * @param result ExpressionResult
	 */
	public void set(final ExpressionResult result) {
		setAll(result.getType(), result.getDouble(), result.getLong(), result.getString(), result.getBool(), result._isState);
	}

	/**
	 * Setzt einen Statuswert
	 *
	 * @param name    Name/Bezeichnung des Statuswerts
	 * @param inRange Ist der Wert innerhalb des Wertebereichs? Falls der Parameter auf true gesetzt ist muss vorher mit {@link #setUnscaledValue(long)} o.ä. eine
	 *                Ganzzahl gesetzt worden sein.
	 */
	public void setState(final String name, final boolean inRange) {
		if(!inRange) {
			// Wert außerhalb des Wertebereichs
			setAll(ResultType.STATUS, 0, 0, name, false, true);
		}
		else {
			if(_type != ResultType.LONG) {
				throw new IllegalStateException("Ein Statuswert innerhalb des Wertebereichs kann nur gesetzt werden wenn ein Integer-Wert gespeichert ist.");
			}
			this._string = name;
			_isState = true;
		}
	}

	/**
	 * Markiert das Objekt als Fehlerhaft.
	 *
	 * @return Gibt ein Objekt mit Typ 'Error' zurück
	 * 
	 * @deprecated Es sollte eine Fehlermeldung angegeben werden
	 */
	@Deprecated
	public static ExpressionResult error() {
		return error(null);
	}

	/**
	 * Markiert das Ergebnis als Fehlerhaft. Eine Fehlerbeschreibung kann mit {@link #getString()} abgerufen werden.
	 *
	 * @param errorMsg Fehlerbeschreibung.
	 *
	 * @return Gibt ein Objekt mit Typ 'Error' zurück
	 */
	public static ExpressionResult error(final String errorMsg) {
		final ExpressionResult r = new ExpressionResult();
		r.set(errorMsg);
		r._type = ResultType.ERROR;
		return r;
	}

	public static ExpressionResult noChange() {
		final ExpressionResult r = new ExpressionResult();
		r._type = ResultType.NO_CHANGE;
		return r;
	}

	/** Markiert das Ergebnis als Fehlerhaft. */
	@Deprecated
	public void setError() {
		this._type = ResultType.ERROR;
	}

	/**
	 * Markiert das Ergebnis als Fehlerhaft.
	 *
	 * @param errorMsg Fehlermeldung.
	 */
	public void setError(final String errorMsg) {
		this._string = errorMsg;
		this._type = ResultType.ERROR;
	}

	/**
	 * Setzt die Werte des ExpressionResult-Objekts.
	 *
	 * @param type        Wert für {@link #_type}
	 * @param doubleValue Wert für {@link #_double}
	 * @param longValue   Wert für {@link #_long}
	 * @param stringValue Wert für {@link #_string}
	 * @param boolValue   Wert für {@link #_bool}
	 * @param isState     Wert für {@link #_isState}
	 */
	private void setAll(
			final ResultType type, final double doubleValue, final long longValue, final String stringValue, final boolean boolValue, final boolean isState) {
		this._type = type;
		this._double = doubleValue;
		this._long = longValue;
		this._string = stringValue;
		this._bool = boolValue;
		this._isState = isState;
		this._conversionFactor = 1.0;
	}


	/**
	 * Gibt den Boolean-Wert dieses ExpressionResult zurück. Genauer gesagt ist die Rückgabe anhängig vom Typ:
	 * <dl>
	 * <dt>BOOL</dt>   <dd>Boolean-Wert</dd>
	 * <dt>LONG</dt>   <dd>false</dd>
	 * <dt>DOUBLE</dt> <dd>false</dd>
	 * <dt>STRING</dt> <dd>false</dd>
	 * <dt>STATUS</dt> <dd>false</dd>
	 * <dt>ERROR</dt>  <dd>false</dd>
	 * <dt>NONE</dt>   <dd>false</dd>
	 * </dl>
	 *
	 * @return Boolean-Wert dieses ExpressionResult
	 */
	public boolean getBool() {
		return _bool;
	}

	/**
	 * Gibt den (unskalierten) Long-Wert dieses ExpressionResult zurück. Genauer gesagt ist die Rückgabe anhängig vom Typ:
	 * <dl>
	 * <dt>BOOL</dt>   <dd>0 (False) oder 1 (True)</dd>
	 * <dt>LONG</dt>   <dd>Gespeicherter Long-Wert</dd>
	 * <dt>DOUBLE</dt> <dd>Mit {@link Math#round(double)} gerundeter Double-Wert</dd>
	 * <dt>STRING</dt> <dd>0</dd>
	 * <dt>STATUS</dt> <dd>0 (Hinweis: Der STATUS-Typ wird nur für Werte <b>außerhalb</b> des Wertebereichs benutzt.)</dd>
	 * <dt>ERROR</dt>  <dd>0</dd>
	 * <dt>NONE</dt>   <dd>0</dd>
	 * </dl>
	 *
	 * @return Long-Wert dieses ExpressionResult
	 */
	public long getLong() {
		return _long;
	}

	/**
	 * Gibt den (skalierten) Double-Wert dieses ExpressionResult zurück. Genauer gesagt ist die Rückgabe anhängig vom Typ:
	 * <dl>
	 * <dt>BOOL</dt>   <dd>0.0 (False) oder 1.0 (True)</dd>
	 * <dt>LONG</dt>   <dd>Gespeicherter Long-Wert in Double konvertiert. PuA-Seitig: Skalierter Wert, Client-Seitig: Unskalierter Wert.</dd>
	 * <dt>DOUBLE</dt> <dd>Gespeicherter Double-Wert</dd>
	 * <dt>STRING</dt> <dd>0.0</dd>
	 * <dt>STATUS</dt> <dd>0.0 (Hinweis: Der STATUS-Typ wird nur für Werte <b>außerhalb</b> des Wertebereichs benutzt.)</dd>
	 * <dt>ERROR</dt>  <dd>0.0</dd>
	 * <dt>NONE</dt>   <dd>0.0</dd>
	 * </dl>
	 *
	 * @return Double-Wert dieses ExpressionResult
	 */
	public double getDouble() {
		return _double;
	}

	/**
	 * Gibt den String-Wert dieses ExpressionResult zurück. Genauer gesagt ist die Rückgabe anhängig vom Typ:
	 * <dl>
	 * <dt>BOOL</dt>   <dd>"true"/"false"</dd>
	 * <dt>LONG</dt>   <dd>Gespeicherter Long-Wert in String konvertiert. PuA-Seitig auch: Der Name des Status-Wert falls der Wert sich innerhalb des Wertebereichs
	 * befindet.</dd>
	 * <dt>DOUBLE</dt> <dd>Gespeicherter Double-Wert mit einem deutschen Zahlenformat in String konvertiert</dd>
	 * <dt>STRING</dt> <dd>Gespeicherter String-Wert</dd>
	 * <dt>STATUS</dt> <dd>Name des gespeicherten Status-Wertes. (Hinweis: Der STATUS-Typ wird nur für Werte <b>außerhalb</b> des Wertebereichs benutzt.)</dd>
	 * <dt>ERROR</dt>  <dd>Falls vorhanden Fehlerbeschreibung, sonst {@code null}.</dd>
	 * <dt>NONE</dt>   <dd>{@code null}</dd>
	 * </dl>
	 *
	 * @return String-Wert dieses ExpressionResult
	 */
	public String getString() {
		return _string;
	}

	/**
	 * Liefert den Typ des Ergebnisses zurück. Siehe {@link ExpressionResult}
	 *
	 * @return Typ des Ergebnisses.
	 */
	public ResultType getType() {
		return _type;
	}


	@Override
	public String toString() {
		if(_type == ResultType.BOOL) {
			return String.valueOf(_bool);
		}
		else if(_type == ResultType.DOUBLE) {
			return formatDouble(_double);
		}
		else if(_type == ResultType.LONG) {
			if(!_isState) return formatLong(_long);
			return formatLong(_long) + " (" + _string + ")";
		}
		else if(_type == ResultType.STRING) {
			return _string;
		}
		else if(_type == ResultType.STATUS) {
			return _string;
		}
		else if(_type == ResultType.ERROR) {
			if(_string != null) {
				return _string;
			}
			else {
				return "Fehler beim Berechnen des Ausdrucks.";
			}
		}
		else if(_type == ResultType.NO_CHANGE){
			return "'";
		}
		else {
			return "-";
		}
	}


	private static String formatDouble(final double v) {
		return _numberFormat.get().format(v);
	}

	private static String formatLong(final long v) {
		return _numberFormat.get().format(v);
	}


	/**
	 * Zeigt an ob ein Ergebnis vorliegt. Prüft dazu ob der Typ ungleich {@code Error} und {@code None} ist. Gibt <code>false</code> zurück, falls {@link
	 * #getType()} <code>none</code> oder <code>error</code> liefert.
	 *
	 * @return <code>true</code> falls ein Ergebnis vorliegt.<code>false</code>sonst.
	 */
	public boolean hasValue() {
		if(_type == ResultType.ERROR || _type == ResultType.NONE) {
			return false;
		}
		else {
			return true;
		}
	}

	/** @see java.lang.Object#equals(java.lang.Object) */
	@Override
	public boolean equals(final Object object) {
		if(object == null) {
			return false;
		}
		else if(object instanceof ExpressionResult) {
			final ExpressionResult other = (ExpressionResult)object;
			if(this.getType() != other.getType()) return false;
			switch(getType()) {
				case BOOL:
					return getBool() == other.getBool();
				case STRING:
					return getString().equals(other.getString());
				case LONG:
					return getLong() == other.getLong();
				case DOUBLE:
					return getDouble() == other.getDouble();
				case STATUS:
					return getString().equals(other.getString());
				case ERROR:
					return getString().equals(other.getString());
				case NONE:
					return true;
				default:
					throw new IllegalStateException("Unbekannter Typ:" + getType());
			}
		}
		return false;
	}

	@Override
	public int hashCode() {
		switch(getType()) {
			case BOOL:
				return getBool() ? 1231 : 1237;
			case STRING:
				return getString().hashCode();
			case LONG:
				return ((Long)getLong()).hashCode();
			case DOUBLE:
				return ((Double)getDouble()).hashCode();
			case STATUS:
				return getString().hashCode();
			case ERROR:
				return getString().hashCode();
			case NONE:
				return 0;
			default:
				throw new IllegalStateException("Unbekannter Typ:" + getType());
		}
	}

	/** @return <code>true</code> falls das Objekt einen Zahlenwert repräsentiert. */
	public boolean isArithmetic() {
		return (getType() == ResultType.LONG || getType() == ResultType.DOUBLE || getType() == ResultType.BOOL);
	}

	/** @see java.lang.Object#clone() */
	@Override
	public ExpressionResult clone() {
		final ExpressionResult clone = new ExpressionResult();
		clone._bool = _bool;
		clone._double = _double;
		clone._long = _long;
		clone._string = _string;
		clone._type = _type;
		clone._isState = _isState;
		return clone;
	}

	public void setNoChange() {
		_type = ResultType.NO_CHANGE;
	}

	public static ExpressionResult state(final String state) {
		ExpressionResult expressionResult = new ExpressionResult();
		expressionResult.setState(state, false);
		return expressionResult;
	}
	
	public static ExpressionResult state(final String state, long value, double conversionFactor) {
		ExpressionResult expressionResult = new ExpressionResult(value, conversionFactor);
		expressionResult.setState(state, true);
		return expressionResult;
	}

	/**
	 * Typ der Werte in {@link ExpressionResult}.
	 *
	 * @author beck et al. projects GmbH
	 * @author Martin Hilgers
	 */
	public enum ResultType {

		/** Bool */
		BOOL,
		/** String */
		STRING,
		/** Long */
		LONG,
		/** Double */
		DOUBLE,
		/** Zustand */
		STATUS,
		/** Ungültiger Wert */
		ERROR,
		/** Kein Wert hinterlegt */
		NONE,
		/** Bei Zustandsprotokollen mit NoChange-Kennzeichnung pro Zelle: Keine Änderung zum vorherigen Datensatz */
		NO_CHANGE
	}
}
