/*
 * Copyright 2015 by Kappich Systemberatung Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.aggregations;

import de.bsvrz.pua.prot.util.ExpressionResult;
import de.bsvrz.pua.prot.util.ResultValue;

import java.util.function.Consumer;
import java.util.function.Supplier;

/**
 * Abstrakte Basisklasse für Aggregationen, die AggregationResultValue zur Vereinfachung benutzen.
 * 
 * Vorgehensweise: AggregationResultValue ableiten, falls Zusatzinformationen je Attribut gespeichert werden sollen (sonst kann
 * AggregationResultValue selbst verwendet werden).
 * 
 * Für die Aggregation diese Klasse ableiten und die erforerlichen Methoden überschreiben. {@linkplain #createAggregationResultValue}
 * sollte dabei eine Methode sein, die eine Instanz des verwendeten AggregationResultValue-Typs erzeugt.
 *
 * Optional kann am Ende finish() überschieben werden um am Ende noch das Ergebnis zu berechnen (z.B. bei der Berechnung des Durchschnitts am Ende noch durch die
 * Anzahl der aggregierten Daten teilen)
 * 
 * @author Kappich Systemberatung
 */
public abstract class AbstractAggregation<T extends AggregationResultValue> implements Aggregation {

	/** Aggregierter Wert */
	private T _result = null;

	/** @see Aggregation#aggregate(de.bsvrz.pua.prot.util.ResultValue) */
	public final void aggregate(ResultValue value) {
		if(_result == null) {
			_result = AggregationResultValue.create(value, createAggregationResultValue(), this);
		}
		else {
			_result.aggregate(value);
		}
	}

	/** @see Aggregation#getResult() */
	public final ResultValue getResult() throws UnsupportedOperationException {
		T temp = _result;
		if(temp == null) {
			// Keine Werte Aggregiert, auf NONE setzen
			temp = createAggregationResultValue().get();
			temp.set(new ExpressionResult());
		}
		temp.finish();
		_result = null;
		return temp;
	}

	protected abstract Supplier<T> createAggregationResultValue();

	/**
	 * Schließt die Berechnung von T als Aggregationswert ab. Hier kann z.B. bei der Durchschnittsaggregation zum Schluss durch die
	 * Anzahl Werte ({@linkplain AggregationResultValue#getAggregationCount()}) dividiert werden.
	 * @param t Bisheriger berechneter Wert und Rückgabe
	 */
	public void finish(final T t) {
	}

	/**
	 * Berechnung der Aggregation, der Implementierende sollte etwas aufrufen wie
	 * <code>
	 *     result.set(x)
	 * </code>
	 * wobei x aus result.getValue() als altem Aggregationswert und newValue als neuem Wert berechnet wird.
	 * @param result Alter Wert und Rückgabe
	 * @param newValue Wert der dazu aggregiert werden soll
	 */
	protected abstract void aggregate(T result, ExpressionResult newValue);

	@Override
	public String toString() {
		return getAggregationName();
	}
}
