/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.client;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.pua.prot.exceptions.ScopeException;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.manager.ProtocolRelatedOperation;
import de.bsvrz.pua.prot.manager.ScriptRelatedOperation;
import de.bsvrz.pua.prot.util.JobInProgress;
import de.bsvrz.pua.prot.util.MemberCheck;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.pua.prot.util.ProtocolRequestResult;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.dynobj.DynObjektException;
import de.bsvrz.sys.funclib.dynobj.DynamischeObjekte;
import de.bsvrz.sys.funclib.losb.datk.*;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.exceptions.SenderException;
import de.bsvrz.sys.funclib.losb.kernsoftware.ConnectionManager;
import de.bsvrz.sys.funclib.losb.kernsoftware.SimpleSender;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;
import de.bsvrz.sys.funclib.losb.util.Tuple;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.GregorianCalendar;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.atomic.AtomicLong;

/**
 * Schnittstelle des Clients zu Protokolle und Auswertungen. Vereinfacht das das Stellen von Anfragen an Pua.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
@SuppressWarnings("deprecation")
public class PuaClient implements AutoCloseable {

	/** Unterstütze Protokoll-Version des Clients */
	public static final int CURRENT_PROTOCOL_VERSION = 4;

	/** Debug Ausgabe */
	private static final Debug debug = Debug.getLogger();

	/** Status der Protokollerstellung konnte nicht abgefragt werden. */
	public static final byte INVALID_STATUS = -1;

	/** Wartezeit in Millisekunden bevor erneut gesendet wird. */
	private static final long WAIT = 500;
	
	/** Sendesteuerung für Protokollanfragen */
	private final PuaSendControlListener _puaSendControlListener;

	/** Verbindung zum Datenverteiler. */
	private final ClientDavInterface _dav;

	/** Datenmodell */
	private final DataModel _model;

	/** AnfrageId. Streng monoton steigend. */
	private static final AtomicLong _requestId_dontUseDirectly = new AtomicLong(1);

	/** Konfigurationsverantwortlicher, an den die Anfrage gesendet wird. */
	private final ConfigurationAuthority _configAuth;

	/** Datenbeschreibung der Protokoll-bezogenen Anfrage. */
	private final DataDescription _ddProtocol;

	/** Datenbeschreibung der Skript-bezogenen Anfrage. */
	private final DataDescription _ddScript;

	/** Objekt an das Pua die Antwort sendet. */
	private final ClientApplication _client;

	/**
	 * Zuletzt automatisch angelegtes Skriptobjekt. Wird von {@link #createScript(DynamicObject,AtlBase,String)} und {@link #createNewScriptObject(String,String)}
	 * gesetzt.
	 */
	public DynamicObject recentlyCreatedScript = null;

	/** Maximale Wartezeit in ms, bis  eine Antwort eingetroffen sein muss bzw. wie lange ein Sendevorgang dauern darf. Default: 60s */
	private long _timeout = 60 * 1000;

	/** Anfrage-Id des letzten Auftrags zur Protokollerstellung. Wird durch {@link #createProtocol(ProcessingParameter)} gesetzt. */
	public long recentRequestId = -1;

	private static final long PROTOCOL_COMPLETION_INTERFACE_ID = -1;

	private final Set<ProtocolListenerEx> _protocolListeners = new HashSet<>();
	
	private final Set<StatusListener> _statusListeners = new CopyOnWriteArraySet<>();

	@SuppressWarnings("unused")
	private final ProtocolAnswerReceiver _protocolAnswerReceiver;

	/**
	 * Größe des Sliding Window für automatische Flusskontrolle beim Protokollabruf
	 */
	private int _flowControlWindowSize = 100;

	private final ExecutorService _flowControlExcecutor = Executors.newSingleThreadExecutor();

	/**
	 * Zustand der Sendesteuerung für PuA-Anfragen (wird nebnutzt um festzustellen, ob PuA läuft)
	 */
	private boolean _listenerOnline = false;
	/**
	 * Von PuA über den Listener-Mechanismus publizierter Archivsystem-Zustand (true == erreichbar, false == nicht erreichbar), false falls PuA nicht läuft.
	 */
	private boolean _archiveOnline = false;

	/**
	 * Ob der PuaClient geschlossen ist
	 */
	private boolean _closed = false;

	/**
	 * Ob eine Sendesteuerungsnachricht für die Protokollanfrage-Attributgruppe schon angekommen ist
	 */
	private boolean _listenerRequested = false;
	
	private int _protocolVersion = 1;

	/**
	 * @param dav        Verbindung zum Datneverteiler.
	 * @param configAuth Konfigurationsverantwortlicher, der verwendet wird.
	 *
	 * @throws ScopeException         Anfrage wird vom Konfigurationsverantwortlichen nicht unterstützt.
	 * @throws ConfigurationException Fehler bei der Kommunikation mit der Konfiguration
	 */
	public PuaClient(ClientDavInterface dav, ConfigurationAuthority configAuth) throws ScopeException, ConfigurationException {
		this._dav = dav;
		_model = dav.getDataModel();

		//Überprüfen, ob der Konfigurationsverantwortliche die Anfrage unterstützt:
		this._configAuth = configAuth;
		MemberCheck.isAttributeGroupMember(configAuth, _model.getAttributeGroup(PidProtocol.atgProtocolRequest));

		_ddProtocol = new DataDescription(_model.getAttributeGroup(PidProtocol.atgProtocolRequest), _model.getAspect(PidProtocol.aspRequest));
		_ddScript = new DataDescription(_model.getAttributeGroup(PidScript.atgRequest), _model.getAspect(PidScript.aspIs));

		_client = dav.getLocalApplicationObject();

		_protocolAnswerReceiver = new ProtocolCompletionAnswerReceiver(_dav, _client, PROTOCOL_COMPLETION_INTERFACE_ID, this);
		_protocolAnswerReceiver.subscribe(ReceiverRole.drain());

		try {
			_puaSendControlListener = new PuaSendControlListener();
			ConnectionManager.subscribeSender(dav, _puaSendControlListener, configAuth, _ddProtocol, SenderRole.sender());
		}
		catch(OneSubscriptionPerSendData oneSubscriptionPerSendData) {
			// Sollte nicht passieren, da Anmeldung als Sender
			throw new IllegalStateException("Verbindung mit PuA kann nicht aufgebaut werden", oneSubscriptionPerSendData);
		}
		synchronized(_puaSendControlListener){
			if(!_listenerRequested) {
				try {
					// Auf erste Sendesteuerung warten, damit beim Start nicht immer fehlerhafterweise PuA als nicht verbunden angezeigt wird 
					_puaSendControlListener.wait(5000);
				}
				catch(InterruptedException ignored) {
				}
			}
		}
	}

	/**
	 * Maximale Wartezeit in ms, bis  eine Antwort eingetroffen sein muss. Default: 60s.
	 *
	 * @param timeout Maximale Wartezeit, bis  eine Antwort eingetroffen sein muss.
	 */
	public void setTimeout(long timeout) {
		this._timeout = timeout;
	}


	/**
	 * Liefert eine eindeutige Anfrage-Id.
	 *
	 * @return Eindeutige Anfrage-Id.
	 */
	private static long getRequestId() {
		// Das & 0xFFFFFFF dient dazu die Request-IDs auf einen Bereich zu beschränken, der Kollisionsfrei mit
		// Protokoll-IDs und Anfrage-IDs ist. Bei protokollbezogenen Operationen wird mit dem ersten Datensatz von der
		// Request-ID zur Kommunikation auf eine PuA-Seitig generierte Anfrage-ID (Job-ID) gewechselt,
		// d.h. Kollisionen müssen hier vermieden werden.
		return _requestId_dontUseDirectly.getAndIncrement() & 0xFFFFFFF;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////
	//									PROTOKOLLTBEZOGENE ANFRAGEN										//
	//////////////////////////////////////////////////////////////////////////////////////////////////	

	/**
	 * Stößt die Protokollerstellung an. Fängt alle(!) Exceptions ab.
	 *
	 * @param pp Auftragsparameter
	 *
	 * @return Objekt zum Abfragen der Antwort.
	 *
	 * @throws FailureException Fehler beim Absetzen der Anfrage.
	 */
	public ProtocolResultStream createProtocol(ProcessingParameter pp) throws FailureException {
		//Eindeutige Id holen
		long id = recentRequestId = getRequestId();
		ProtocolAnswerReceiver receiver = null;
		ProtocolResultStream prs = null;
		try {
			// schon vorher auf Antwort anmelden
			receiver = new ProtocolAnswerReceiver(_dav, _client, id, true, _timeout);
			receiver.subscribe(ReceiverRole.drain());
			// Daten serialisieren & senden:
			final ProtocolRequestResult result = sendRequest(id, ProtocolRelatedOperation.CREATE, pp.serialize());
			prs = new ProtocolResultStream(receiver, result, this);
			receiver.waitForAnswer(_timeout);
		}
		catch(ConfigurationException e) {
			prs = new ProtocolResultStream(null, new ProtocolRequestResult(ErrorMessage.COMMUNICATION + e.getMessage()), this);
		}
		catch(Exception e) {
			prs = new ProtocolResultStream(null, new ProtocolRequestResult(e), this);
		}
		return prs;
	}

	/**
	 * Meldet einen Empfänger an, der über fertiggestellte Protokolle informiert wird
	 *
	 * @param listener Empfänger
	 * @return Rückmeldung auf die Anfrage zur Prüfung ob die Operation fehlerfrei durchgeführt wurde
	 * @throws FailureException Exception
	 */
	public ProtocolRequestResult addProtocolListener(final ProtocolListenerEx listener) throws FailureException {
		if(listener == null) throw new IllegalArgumentException("listener ist null");
		ProtocolRequestResult result = new ProtocolRequestResult();
		synchronized(_protocolListeners) {
			_protocolListeners.add(listener);
		}
		return result;
	}

	/**
	 * Meldet einen Empfänger ab, der über fertiggestellte Protokolle informiert wurde
	 * @param listener Empfänger
	 * @return Rückmeldung auf die Anfrage zur Prüfung ob die Operation fehlerfrei durchgeführt wurde
	 * @throws FailureException Exception
	 */
	public ProtocolRequestResult removeProtocolListener(final ProtocolListenerEx listener) throws FailureException {
		if(listener == null) throw new IllegalArgumentException("listener ist null");
		ProtocolRequestResult result = new ProtocolRequestResult();
		synchronized(_protocolListeners) {
			_protocolListeners.remove(listener);
		}
		return result;
	}

	/**
	 * Meldet einen Empfänger an, der über fertiggestellte Protokolle informiert wird
	 *
	 * @param listener Empfänger. Für erweiterte Benachrichtigungen kann und sollte hier eine Instanz von {@link
	 *                 de.bsvrz.pua.prot.client.ProtocolListenerEx} statt {@link de.bsvrz.pua.prot.client.ProtocolListener} verwendet
	 *                 werden.
	 * @return Rückmeldung auf die Anfrage zur Prüfung ob die Operation fehlerfrei durchgeführt wurde
	 * @throws FailureException Exception
	 */
	public ProtocolRequestResult addProtocolListener(final ProtocolListener listener) throws FailureException {
		if(listener == null) throw new IllegalArgumentException("listener ist null");
		return addProtocolListener(new ProtocolListenerWrapper(listener));
	}

	/**
	 * Fügt einen Listener hinzu, mit dem sich der Anwender über Änderungen am Verfügbarkeitszustand von PuA 
	 * und dem von PuA verwendeten Archivsystem informieren kann. Der Listener wird direkt nach dem Anmelden über den aktuellen Zustand informiert. 
	 * @param listener Listener-Interface (nicht null)
	 * @return true wenn der Listener hinzugefügt wurde. False: Der Listener war bereits angemeldet.
	 */
	public boolean addStatusListener(final StatusListener listener)  {
		if(listener == null) throw new IllegalArgumentException("listener ist null");
		if(_statusListeners.add(listener)){
			listener.statusChanged(isPuaOnline(), isArchiveOnline());
			return true;
		}
		return false;
	}

	/**
	 * Entfernt einen, mit dem sich der Anwender über Änderungen am Verfügbarkeitszustand von PuA 
	 * und dem von PuA verwendeten Archivsystem informieren kann
	 * @param listener Listener-Interface (nicht null)
	 * @return true wenn der Listener entfernt wurde. False: Der Listener war noch nicht angemeldet.
	 */
	public boolean removeStatusListener(final StatusListener listener)  {
		if(listener == null) throw new IllegalArgumentException("listener ist null");
		return _statusListeners.remove(listener);
	}

	/**
	 * Meldet einen Empfänger ab, der über fertiggestellte Protokolle informiert wurde
	 * @param listener Empfänger
	 * @return Rückmeldung auf die Anfrage zur Prüfung ob die Operation fehlerfrei durchgeführt wurde
	 * @throws FailureException Exception
	 */
	public ProtocolRequestResult removeProtocolListener(final ProtocolListener listener) throws FailureException {
		if(listener == null) throw new IllegalArgumentException("listener ist null");
		return removeProtocolListener(new ProtocolListenerWrapper(listener));
	}

	private ProtocolRequestResult startStopProtocolListener(final byte... action) throws FailureException {
		return sendRequest(
				PROTOCOL_COMPLETION_INTERFACE_ID, ProtocolRelatedOperation.PROTOCOL_COMPLETION_INTERFACE, action
		);
	}

	/**
	 * Fragt ein bereits gespeichertes Protokoll ab.
	 *
	 * @param protocolId Protokoll-Id des gespeicherten Protokolls.
	 *
	 * @return Objekt zum Abfragen der Antwort.
	 *
	 * @throws FailureException Fehler beim Absetzen der Anfrage.
	 */
	public ProtocolResultStream getSavedProtocol(long protocolId) throws FailureException {
		return getProtocol(protocolId, true);
	}

	/**
	 * Fragt ein noch nicht abgerufenes gespeichertes Protokoll ab.
	 *
	 * @param protocolId Protokoll-Id des gespeicherten Protokolls.
	 *
	 * @return Objekt zum Abfragen der Antwort.
	 *
	 * @throws FailureException Fehler beim Absetzen der Anfrage.
	 */
	public ProtocolResultStream getUnreadProtocol(long protocolId) throws FailureException {
		return getProtocol(protocolId, false);
	}

	/**
	 * Fragt ein bereits gespeichertes bzw. noch nicht abgerufenens Protokoll ab.
	 *
	 * @param protocolId Protokoll-Id des gespeicherten Protokolls.
	 * @param read       <code>true</code> gespeicherte Protokolle abrufen. <code>false</code> noch nicht abgerufene Protokolle senden.
	 *
	 * @return Objekt zum Abfragen der Antwort.
	 *
	 * @throws FailureException Fehler beim Absetzen der Anfrage.
	 */
	public ProtocolResultStream getProtocol(long protocolId, boolean read) throws FailureException {
		//Eindeutige Id holen
		long id = getRequestId();
		ProtocolAnswerReceiver receiver = null;
		ProtocolResultStream prs = null;
		try {
			//schon vorher auf die Antwort anmelden
			receiver = new ProtocolAnswerReceiver(_dav, _client, id, true, _timeout);
			receiver.subscribe(ReceiverRole.drain());
			//Daten serialisieren und senden
			ProtocolRequestResult result = sendRequest(
					id,
					(read) ? ProtocolRelatedOperation.SEND_SAVED_PROTOCOL : ProtocolRelatedOperation.SEND_UNREAD_PROTOCOL,
					SerializerUtil.serializeId(protocolId)
			);
			prs = new ProtocolResultStream(receiver, result, this);
			receiver.waitForAnswer(_timeout);
		}
		catch(ConfigurationException e) {
			prs = new ProtocolResultStream(null, new ProtocolRequestResult(ErrorMessage.COMMUNICATION + e.getMessage()), this);
		}
		catch(Exception e) {
			prs = new ProtocolResultStream(null, new ProtocolRequestResult(e), this);
		}
		return prs;
	}

	/**
	 * Stellt eine Anfrage an Protokolle und Auswertungen. Wartet, bis der Auftrag abgesendet wurde.
	 *
	 * @param requestId   AnfrageId
	 * @param opCode      Operationscode
	 * @param requestData Auftragsparameter
	 *
	 * @return Status der Anfrage. Falls Anfrage Fehlschlägt, dann lag ein Fehler beim Senden vor.
	 *
	 * @throws FailureException Fehler bei Kommunikation mit der Konfiguration.
	 */
	private ProtocolRequestResult sendRequest(long requestId, int opCode, byte[] requestData) throws FailureException {
		if(_closed) throw new IllegalStateException("close()-Methode wurde schon aufgerufen");
		try {
			//Zusammenstellen der Anfrage & senden:
			Data request = AtgProtocolRequest.build(_dav, _client, requestId, opCode, requestData);
			try{
				request.getUnscaledValue("Flusskontrolle").set(_flowControlWindowSize * 2);
			}
			catch(Exception e){
				debug.fine("Datenmodell unterstützt keine Flusskontrolle", e);
			}
			if(!request.isDefined()) {
				return new ProtocolRequestResult(ErrorMessage.OLD_DATAMODEL);
			}
			if(SimpleSender.sendWait(_dav, _configAuth, _ddProtocol, request, _timeout)) {
				return new ProtocolRequestResult();
			}
			else {
				return new ProtocolRequestResult(ErrorMessage.CAN_NOT_SEND);
			}
		}
		catch(SenderException e) {
			return new ProtocolRequestResult(e.getMessage());
		}
	}

	/**
	 * Hält eine protokollbezogene Operation an. Gültige Operationen für diese Methode sind:
	 *
	 * <ul>
	 *     <li>Erstellung von Protokollen
	 *     <li>Abruf von gespeicherten Protokollen
	 *     <li>Abruf von ungelesenen Protokollen
	 * </ul>
	 *
	 * @param jobId Auftrags-ID des anzuhaltenden Auftrags. Bei Protokollerstellungsaufträgen identisch mit Protokoll-ID.
	 *              Die Auftrags-ID kann über {@link ProtocolResultStream#getJobId()}
	 *              oder {@link #getJobList(de.bsvrz.pua.prot.util.ProtocolRequestResult)} abgerufen werden.
	 *
	 * @return Informationen über das Absetzen der Anfrage.
	 *
	 * @throws FailureException Fehler bei der Kommunikation mit der Konfiguration.
	 * @deprecated Diese Methode war zur Flusskontrolle gedacht und wird nicht mehr benötigt, weil mittlerweile eine automatische Flusskontrolle zwischen
	 *             PuA-Server und -Client implementiert ist.
	 */
	@Deprecated
	public ProtocolRequestResult suspendProtocol(long jobId) throws FailureException {
		long requestId = getRequestId();
		//Daten serialisieren & senden:
		return sendRequest(requestId, ProtocolRelatedOperation.PAUSE_PROTOCOL, SerializerUtil.serializeId(jobId));
	}

	/**
	 * Setzt eine protokollbezogene Operation fort. Gültige Operationen für diese Methode sind:
	 *
	 * <ul>
	 *     <li>Erstellung von Protokollen
	 *     <li>Abruf von gespeicherten Protokollen
	 *     <li>Abruf von ungelesenen Protokollen
	 * </ul>
	 *
	 * @param jobId Auftrags-ID des fortzusetzenden Auftrags. Bei Protokollerstellungsaufträgen identisch mit Protokoll-ID.
	 *              Die Auftrags-ID kann über {@link ProtocolResultStream#getJobId()}
	 *              oder {@link #getJobList(de.bsvrz.pua.prot.util.ProtocolRequestResult)} abgerufen werden.
	 *
	 * @return Informationen über das Absetzen der Anfrage.
	 *
	 * @throws FailureException Fehler bei der Kommunikation mit der Konfiguration.
	 * @deprecated Diese Methode war zur Flusskontrolle gedacht und wird nicht mehr benötigt, weil mittlerweile eine automatische Flusskontrolle zwischen
	 *             PuA-Server und -Client implementiert ist.
	 */
	@Deprecated
	public ProtocolRequestResult resumeProtocol(long jobId) throws FailureException {
		long requestId = getRequestId();
		//Daten serialisieren & senden:
		return sendRequest(requestId, ProtocolRelatedOperation.RESUME_PROTOCOL, SerializerUtil.serializeId(jobId));
	}

	/**
	 * Bricht eine protokollbezogene Operation ab. Gültige Operationen für diese Methode sind:
	 *
	 * <ul>
	 *     <li>Erstellung von Protokollen
	 *     <li>Abruf von gespeicherten Protokollen
	 *     <li>Abruf von ungelesenen Protokollen
	 * </ul>
	 *
	 * {@link ProtocolResultStream#take()} liefert nach dem Abbrechen <code>null</code> und {@link ProtocolResultStream#hasNext()} <code>true</code> zurück.
	 *
	 * @param jobId Auftrags-ID des abzubrechenden Auftrags. Bei Protokollerstellungsaufträgen identisch mit Protokoll-ID.
	 *              Die Auftrags-ID kann über {@link ProtocolResultStream#getJobId()}
	 *              oder {@link #getJobList(ProtocolRequestResult)} abgerufen werden.
	 *
	 * @return Informationen über das Absetzen der Anfrage. Positive Quittung nur dann, wenn die Protokollerstellung abgebrochen wurde.
	 */
	public ProtocolRequestResult abortProtocol(long jobId) {
		return abortProtocol(jobId, false);
	}

	/**
	 * Bricht eine protokollbezogene Operation ab. Gültige Operationen für diese Methode sind:
	 *
	 * <ul>
	 *     <li>Erstellung von Protokollen
	 *     <li>Abruf von gespeicherten Protokollen
	 *     <li>Abruf von ungelesenen Protokollen
	 * </ul>
	 *
	 * {@link ProtocolResultStream#take()} liefert nach dem Abbrechen <code>null</code> und {@link ProtocolResultStream#hasNext()} <code>true</code> zurück.
	 *
	 * @param jobId Auftrags-ID des abzubrechenden Auftrags. Bei Protokollerstellungsaufträgen identisch mit Protokoll-ID.
	 *              Die Auftrags-ID kann über {@link ProtocolResultStream#getJobId()}
	 *              oder {@link #getJobList(ProtocolRequestResult)} abgerufen werden.
	 *
	 * @param resumeSaving Falls es sich um einen Auftrag zur Protokollerstellung handelt: Soll das Protokoll trotzdem gespeichert werden? Es wird dann nur die
	 *                     Übertragung des Protokolls abgebrochen. True wird nur von PuA-Versionen ab 2.9.0 unterstützt.   
	 * 
	 * @return Informationen über das Absetzen der Anfrage. Positive Quittung nur dann, wenn die Protokollerstellung abgebrochen wurde.
	 */
	public ProtocolRequestResult abortProtocol(long jobId, final boolean resumeSaving) {
		long requestId = getRequestId();
		ProtocolAnswerReceiver receiver = null;
		//Auf Antwort anmelden
		try {
			receiver = new ProtocolAnswerReceiver(_dav, _client, requestId, false, _timeout);
			receiver.subscribe(ReceiverRole.drain());
			ProtocolRequestResult result = null;
			result = sendRequest(requestId, resumeSaving ? ProtocolRelatedOperation.ACK_FLAG : ProtocolRelatedOperation.ABORT, SerializerUtil.serializeId(jobId));
			if(result.isRequestSuccessful()) {
				ResultData rd = receiver.take();
				if(rd == null) {
					result.set(ErrorMessage.TIMEOUT);
				}
				else {
					AtgProtocolRequest answer = AtgProtocolRequest.getJavaObject(rd.getData());
					if(answer.opCode != ProtocolRelatedOperation.ACK_FLAG) {
						try {
							result.set((String)SerializerUtil.deserialize(answer.requestData));
						}
						catch(Exception e) {
							result.set(ErrorMessage.UNEXPECTED_DATA_TYPE + SerializerUtil.deserialize(answer.requestData).getClass().getName());
						}
					}
				}
			}
			return result;
		}
		catch(FailureException e) {
			return new ProtocolRequestResult(e.getMessage());
		}
		catch(ConfigurationException e) {
			return new ProtocolRequestResult(ErrorMessage.COMMUNICATION + e.getMessage());
		}
		catch(InterruptedException e) {
			return new ProtocolRequestResult(ErrorMessage.INTERRUPTED_EXCEPTION + e.getMessage());
		}
		finally {
			if(receiver != null) receiver.unsubscribe();
		}
	}

	/**
	 * Liefert den Status der Protokollerstellung.
	 *
	 * @param jobId Auftrags-ID des abzurufenden Auftrags. Bei Protokollerstellungsaufträgen identisch mit Protokoll-ID.
	 *              Die Auftrags-ID kann über {@link ProtocolResultStream#getJobId()}
	 *              oder {@link #getJobList(de.bsvrz.pua.prot.util.ProtocolRequestResult)} abgerufen werden.
	 * @param resultInfo Informationen über den Ablauf der Aktion. Kann <code>null</code> sein.
	 *
	 * @return Status der Protokollerstellung. -1 bei Fehler.
	 */
	public byte getStatus(long jobId, ProtocolRequestResult resultInfo) {
		if(resultInfo == null) resultInfo = new ProtocolRequestResult();
		long requestId = getRequestId();
		ProtocolAnswerReceiver receiver = null;
		//Auf Antwort anmelden
		try {
			receiver = new ProtocolAnswerReceiver(_dav, _client, requestId, false, _timeout);
			receiver.subscribe(ReceiverRole.drain());
			resultInfo.set(sendRequest(requestId, ProtocolRelatedOperation.STATUS, SerializerUtil.serializeId(jobId)));
			if(resultInfo.isRequestSuccessful()) {
				ResultData rd = receiver.take();
				if(rd == null) {
					resultInfo.set(ErrorMessage.TIMEOUT);
					return INVALID_STATUS;
				}
				AtgProtocolRequest answer = AtgProtocolRequest.getJavaObject(rd.getData());
				if(answer.opCode == ProtocolRelatedOperation.STATUS && answer.requestData.length == 1) {
					return answer.requestData[0];
				}
				try {
					resultInfo.set((String)SerializerUtil.deserialize(answer.requestData));
				}
				catch(Exception e) {
					resultInfo.set(ErrorMessage.UNEXPECTED_DATA_TYPE + SerializerUtil.deserialize(answer.requestData).getClass().getName());
				}
				return INVALID_STATUS;
			}
			return INVALID_STATUS;
		}
		catch(Exception e) {
			e.printStackTrace();
			resultInfo.set(e.getMessage());
			return INVALID_STATUS;
		}
		finally {
			if(receiver != null) receiver.unsubscribe();
		}
	}

	private boolean getIsActiveOld(ProtocolRequestResult resultInfo) {
		if(resultInfo == null) resultInfo = new ProtocolRequestResult();
		long requestId = getRequestId();
		ProtocolAnswerReceiver receiver = null;
		//Auf Antwort anmelden
		try {
			receiver = new ProtocolAnswerReceiver(_dav, _client, requestId, false, _timeout);
			receiver.subscribe(ReceiverRole.drain());
			resultInfo.set(sendRequest(requestId, ProtocolRelatedOperation.PUA_ARS_STATUS, new byte[0]));
			if(resultInfo.isRequestSuccessful()) {
				ResultData rd = receiver.take();
				if(rd == null) {
					resultInfo.set(ErrorMessage.TIMEOUT);
					return false;
				}
				final AtgProtocolRequest answer = AtgProtocolRequest.getJavaObject(rd.getData());
				return answer.opCode == ProtocolRelatedOperation.PUA_ARS_STATUS && answer.requestData.length == 1 && answer.requestData[0] == 1;
			}
			return false;
		}
		catch(Exception e) {
			resultInfo.set(e.getMessage());
			return false;
		}
		finally {
			if(receiver != null) receiver.unsubscribe();
		}
	}
	
	/**
	 * Prüft ob PuA aktiv ist und das Archivsystem erreichbar ist, also ob Protokollanfragen gestellt werden können.
	 *
	 * @param resultInfo Ergebnis der Anfrage, seit PuA 2.9.0 immer erfolgreich
	 *
	 * @return true wenn  alles in Ordnung ist und Anfragen gestellt werden können, false wenn PuA nicht erreichbar ist oder das Archivsystem von PuA nicht erreichbar ist
	 * 
	 * @deprecated  Kann durch die Funktionen {@linkplain #isPuaOnline} und {@linkplain #isArchiveOnline} ersetzt werden, die detaillierte Informationen bieten
	 */
	@Deprecated
	public boolean getIsActive(ProtocolRequestResult resultInfo) {
		if(resultInfo != null) resultInfo.setSuccessful();
		return isArchiveOnline();
	}

	/** 
	 * Gibt {@code true} zurück, wenn PuA erreichbar ist
	 * @return {@code true}, wenn PuA erreichbar ist, sonst {@code false}
	 */
	public boolean isPuaOnline() {
		return _listenerOnline;
	}

	/**
	 * Gibt {@code true} zurück, wenn PuA erreichbar ist und PuA das Archiv erreichen kann. Wenn PuA nicht läuft bzw. nicht erreichbar ist, gibt diese Methode
	 * immer {@code false} zurück, da diese Methode prüft, ob PuA das von PuA verwendete Archivsystem erreichen kann und nicht, ob irgendein Archivsystem
	 * läuft.
	 *
	 * @return {@code true}, wenn PuA erreichbar ist und PuA das Archiv erreichen kann, sonst {@code false}
	 */
	public boolean isArchiveOnline() {
		return _archiveOnline;
	}

	/**
	 * Liefert eine Liste der gespeicherten Protokolle und Protokollinformationen.
	 *
	 * @param resultInfo In diesem Objekt werden Fehlermeldungen abgelegt.
	 *
	 * @param user Benutzer, für den die Anfrage ausgeführt wird. (kann null oder ein Leerstring sein, dann werden keine benutzerspezifischen Skripte zurückgegeben)
	 * @return Array mit Ids der gespeicherten Protokolle oder <code>null</code> bei Fehler.
	 */
	public Tuple<Long, ProcessingParameter>[] getSavedProtocolList(ProtocolRequestResult resultInfo, final String user) {
		return getProtocolList(resultInfo, user, true);
	}

	/**
	 * Liefert eine Liste der noch nicht abgerufenen Protokolle und Protokollinformationen.
	 *
	 * @param resultInfo In diesem Objekt werden Fehlermeldungen abgelegt.
	 *
	 * @param user Benutzer, für den die Anfrage ausgeführt wird. (kann null oder ein Leerstring sein, dann werden keine benutzerspezifischen Protokolle zurückgegeben)
	 * @return Array mit Ids der gespeicherten Protokolle oder <code>null</code> bei Fehler.
	 */
	public Tuple<Long, ProcessingParameter>[] getUnreadProtocolList(ProtocolRequestResult resultInfo, final String user) {
		return getProtocolList(resultInfo, user, false);
	}

	/**
	 * Liefert eine Liste der gespeicherten bzw. noch nicht abgerufenen Protokolle und Protokollinformationen.
	 *
	 * @param resultInfo In diesem Objekt werden Fehlermeldungen abgelegt.
	 * @param user Benutzer, für den die Anfrage ausgeführt wird. (kann null oder ein Leerstring sein, dann werden keine benutzerspezifischen Protokolle zurückgegeben)
	 * @param read       <code>true</code> gespeicherte Protokolle abrufen. <code>false</code> noch nicht abgerufene Protokolle senden.
	 *
	 * @return Array mit Ids der gespeicherten Protokolle oder <code>null</code> bei Fehler.
	 */
	private Tuple<Long, ProcessingParameter>[] getProtocolList(ProtocolRequestResult resultInfo, String user, boolean read) {
		if(resultInfo == null) resultInfo = new ProtocolRequestResult();

		long requestId = getRequestId();
		ProtocolAnswerReceiver receiver = null;
		AtgProtocolRequest answer = null;
		//Auf Antwort anmelden
		try {
			receiver = new ProtocolAnswerReceiver(_dav, _client, requestId, false, _timeout);
			receiver.subscribe(ReceiverRole.drain());
			resultInfo.set(
					sendRequest(
							requestId, (read) ? ProtocolRelatedOperation.LIST : ProtocolRelatedOperation.UNREAD_PROTOCOL_LIST, SerializerUtil.serializeToByteArray(user)
					)
			);
			if(resultInfo.isRequestSuccessful()) {
				ResultData rd = receiver.take();
				if(rd == null) {
					resultInfo.set(ErrorMessage.TIMEOUT);
				}
				else {
					answer = AtgProtocolRequest.getJavaObject(rd.getData());
					if(answer.opCode != ProtocolRelatedOperation.ACK_FLAG) {
						resultInfo.set((String)SerializerUtil.deserialize(answer.requestData));
					}
					else {
						return deserialize(answer.requestData, resultInfo);
					}
				}
			}
		}
		catch(FailureException e) {
			resultInfo.set(e.getMessage());
		}
		catch(ConfigurationException e) {
			resultInfo.set(ErrorMessage.COMMUNICATION + e.getMessage());
		}
		catch(Exception e) {
			resultInfo.set(ErrorMessage.UNEXPECTED_DATA_TYPE + ((answer != null) ? SerializerUtil.deserialize(answer.requestData).getClass().getName() : ""));
		}
		finally {
			if(receiver != null) receiver.unsubscribe();
		}
		return null;
	}

	/**
	 * Deserialisiert Protokollids und Protokollinformationen.
	 *
	 * @param data       Serialisierte Daten.
	 * @param resultInfo Wird mit Ergebnis der Operation belegt.
	 *
	 * @return Liste mit Protokoll-Ids und Protokollinformationen.
	 */
	@SuppressWarnings("unchecked")
	private Tuple<Long, ProcessingParameter>[] deserialize(byte[] data, ProtocolRequestResult resultInfo) {
		ObjectInputStream ois = null;
		try {
			ByteArrayInputStream bis = new ByteArrayInputStream(data);
			ois = new ObjectInputStream(bis);
			long[] protocolIds = null;
			ProcessingParameter[] pp;

			Object o = ois.readObject();
			if(o instanceof byte[][]) {
				byte[][] field = (byte[][])o;
				o = SerializerUtil.deserialize(field[0]);
				if(o instanceof long[]) {
					protocolIds = (long[])o;
					pp = new ProcessingParameter[protocolIds.length];
					for(int i = 1; i < field.length; i++) {
						pp[i - 1] = ProcessingParameter.deserialize(_dav.getDataModel(), field[i]);
					}
					Tuple<Long, ProcessingParameter>[] result = new Tuple[protocolIds.length];
					for(int i = 0; i < protocolIds.length; i++) {
						result[i] = new Tuple<>(protocolIds[i], pp[i]);
					}

					resultInfo.setSuccessful();
					return result;
				}
				else {
					return setAndLog(resultInfo, ErrorMessage.UNEXPECTED_DATA_TYPE + o.getClass().getName() + " <-> long[] in PuaClient.deserialize()");
				}
			}
			else {
				return setAndLog(resultInfo, ErrorMessage.UNEXPECTED_DATA_TYPE + o.getClass().getName() + " <-> byte[][] in PuaClient.deserialize()");
			}
		}
		catch(ClassNotFoundException e) {
			return setAndLog(resultInfo, ErrorMessage.CAN_NOT_DESERIALIZE + e.getMessage());
		}
		catch(IOException e) {
			return setAndLog(resultInfo, ErrorMessage.CAN_NOT_ACCESS_FILE + e.getMessage());
		}
		catch(FailureException e) {
			return setAndLog(resultInfo, e.getMessage());
		}
		finally {
			Util.close(ois);
		}
	}


	/**
	 * Fehlerbehandlung
	 *
	 * @param resultInfo
	 * @param string
	 *
	 * @return Immer null.
	 */
	private Tuple<Long, ProcessingParameter>[] setAndLog(ProtocolRequestResult resultInfo, String string) {
		resultInfo.set(string);
		debug.warning(string);
		return null;
	}


	/**
	 * Löscht ein gespeichertes Protokoll.
	 *
	 * @param protocolId Protokoll das gelöscht werden soll.
	 *
	 * @return Informationen über den Ausgang des Löschvorgangs.
	 */
	public ProtocolRequestResult deleteSaveProtocol(long protocolId) {
		ProtocolRequestResult resultInfo = new ProtocolRequestResult();
		long requestId = getRequestId();
		ProtocolAnswerReceiver receiver = null;
		//Auf Antwort anmelden
		try {
			receiver = new ProtocolAnswerReceiver(_dav, _client, requestId, false, _timeout);
			receiver.subscribe(ReceiverRole.drain());
			resultInfo = sendRequest(requestId, ProtocolRelatedOperation.DELETE, SerializerUtil.serializeId(protocolId));
			if(resultInfo.isRequestSuccessful()) {
				ResultData rd = receiver.take();
				if(rd == null) {
					resultInfo.set(ErrorMessage.TIMEOUT);
				}
				else {
					AtgProtocolRequest answer = AtgProtocolRequest.getJavaObject(rd.getData());
					try {
						if(answer.opCode != ProtocolRelatedOperation.ACK_FLAG) resultInfo.set((String)SerializerUtil.deserialize(answer.requestData));
					}
					catch(Exception e) {
						resultInfo.set(ErrorMessage.UNEXPECTED_DATA_TYPE + SerializerUtil.deserialize(answer.requestData).getClass().getName());
					}
				}
			}
		}
		catch(FailureException e) {
			resultInfo.set(e.getMessage());
		}
		catch(ConfigurationException e) {
			resultInfo.set(ErrorMessage.COMMUNICATION + e.getMessage());
		}
		catch(InterruptedException e) {
			return new ProtocolRequestResult(ErrorMessage.INTERRUPTED_EXCEPTION + e.getMessage());
		}
		finally {
			if(receiver != null) receiver.unsubscribe();
		}
		return resultInfo;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////
	//									SKRIPTBEZOGENE ANFRAGEN										//
	//////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Sendet den Auftrag 'Skript anlegen/editieren'. Wartet auf eine Antwort von PuA. Behandelt alle(!) Exceptions. Wertet nur die erste Antwort aus, und beendet
	 * danach den Empfang von Daten. Es wird zweimal versucht, die Anfrage an PuA zu senden. Dies hat den Grund, dass, falls von createScript auch das dynamische
	 * Skriptobjekt erzeugt wurde, PuA eine gewisse Zeit gegeben werde muss, um sich als Empfänger auf dieses Skriptobjekt anzumelden. Das Skriptobjekt wird zudem
	 * in {@link #recentlyCreatedScript} abgelegt.
	 *
	 * @param pid      Pid des Skriptobjekts, das angelegt / editiert werden soll. Sollte nicht <code>null</code> sein. Existiert noch kein Skriptobjekt mit der
	 *                 übergebenen pid, so wird eins angelegt.
	 * @param name     Name des Skripts.
	 * @param baseInfo Grundinformationen.
	 * @param source   Skript Quelltext.
	 *
	 * @return Ergebnis der Anfrage.
	 */
	public ProtocolRequestResult createScript(String pid, String name, AtlBase baseInfo, String source) {
		SystemObject script = _model.getObject(pid);
		if(script == null)										//ggf Skript anlegen
		{
			script = createNewScriptObject(pid, name);
		}
		else if(!script.getType().equals(_model.getType(PidScript.type)))		//falls Objekt existiert: Ist es vom richtigen Typ?
		{
			return new ProtocolRequestResult("Pid ist schon an ein Objekt vergeben. Dieses Objekt kann nicht verwendet werden, da es kein Skriptobjekt ist.");
		}
		if(script == null)		//Anlegen hat nicht funktioniert.
		{
			return new ProtocolRequestResult("Fehler beim Anlegen des Skripts.");
		}
		return createScript((DynamicObject)script, baseInfo, source);		//Operation durchführen.
	}

	/**
	 * Sendet den Auftrag 'Skript anlegen/editieren'. Wartet auf eine Antwort von PuA. Behandelt alle(!) Exceptions. Wertet nur die erste Antwort aus, und beendet
	 * danach den Empfang von Daten. Es wird zweimal versucht, die Anfrage an PuA zu senden. Dies hat den Grund, dass, falls von createScript auch das dynamische
	 * Skriptobjekt erzeugt wurde, PuA eine gewisse Zeit gegeben werde muss, um sich als Empfänger auf dieses Skriptobjekt anzumelden.
	 *
	 * @param script   Skriptobjekt das angelegt / editiert werden soll. Kann <code>null</code> sein. Dann wird ein neues Skriptobjekt erzeugt. Der Name des
	 *                 erzeugten Skriptobjekts ist das aktuelle Datum + Zeit. Auf das Skriptobjekt kann durch {@link #recentlyCreatedScript} zugegriffen werden.
	 * @param baseInfo Grundinformationen.
	 * @param source   Skript Quelltext.
	 *
	 * @return Ergebnis der Anfrage.
	 */
	public ProtocolRequestResult createScript(DynamicObject script, AtlBase baseInfo, String source) {
		ScriptAnswerReceiver receiver = null;
		try {
			if(script == null) {
				GregorianCalendar gc = new GregorianCalendar();
				gc.clear();
				gc.setTimeInMillis(System.currentTimeMillis());
				script = createNewScriptObject("pua.testSkript." + System.currentTimeMillis(), "Skript vom " + gc.getTime().toString());
				if(script == null) return new ProtocolRequestResult(ErrorMessage.CAN_NOT_CREATE_SCRIPT_OBJECT);
			}
			this.recentlyCreatedScript = script;

			long requestId = getRequestId();
			//schon vorher auf Antwort anmelden
			receiver = new ScriptAnswerReceiver(_dav, script, _client, requestId, _timeout);
			receiver.subscribe(ReceiverRole.receiver());
			
			//Anfrage zusammenstellen und absenden
			AtlExtra atlExtra = new AtlExtra(_client, requestId, ScriptRelatedOperation.CREATE, source);
			Data requestData = AtgScriptRequest.build(_dav, baseInfo, atlExtra);
			boolean sent = SimpleSender.sendWait(_dav, script, _ddScript, requestData, _timeout);
			if(!sent) {
				sleep();
				sent = SimpleSender.sendWait(_dav, script, _ddScript, requestData, _timeout);
			}
			if(sent) {
				//jetzt auf das Ergebnis warten & vergleichen.
				ResultData rd = receiver.take();
				if(rd == null) {
					return new ProtocolRequestResult(ErrorMessage.TIMEOUT);
				}
				else {
					AtgAnswer answer = AtgAnswer.getJavaObject(rd.getData());
					if(answer.atlExtra.opCode == ScriptRelatedOperation.ACK_FLAG) {
						return new ProtocolRequestResult();
					}
					else {
						return new ProtocolRequestResult(answer.message);
					}
				}
			}
			else {
				return new ProtocolRequestResult(ErrorMessage.SENDING_NOT_ALLOWED);
			}
		}
		catch(ConfigurationException e) {
			return new ProtocolRequestResult(ErrorMessage.COMMUNICATION + e.getMessage());
		}
		catch(Exception e) {
			return new ProtocolRequestResult(e.getMessage());
		}
		finally {
			if(receiver != null) receiver.unsubscribe();
		}
	}

	/** Wartet {@link #WAIT} Millisekunden lang. */
	private void sleep() {
		try {
			Thread.sleep(WAIT);
		}
		catch(Exception e) {
			debug.finest(ErrorMessage.INTERRUPTED_EXCEPTION, e);
		}
	}


	/**
	 * Erzeugt ein neues dynamischen Skriptobjekt und fügt es in die Menge der dynamischen Skriptobjekte ein.
	 *
	 * @param pid  Pid die dem Objekt zugewiesen wird. Kann <code>null</code> sein.
	 * @param name Name der dem Objekt zugewiesen wird. Kann <code>null</code> sein.
	 *
	 * @return Erzeugtes Skriptobjekt oder <code>null</code> im Fehlerfall.
	 */
	private DynamicObject createNewScriptObject(String pid, String name) {
		DynamicObject script = null;
		try {
			DynamicObjectType dynamicObjectType = null;
			SystemObjectType sot = _model.getType(PidScript.type);
			if(sot instanceof DynamicObjectType) {
				dynamicObjectType = (DynamicObjectType)sot;
			}
			else {
				throw new FailureException(InterpreterErrorMessage.INVALID_PID, LoggerException.OTHER);
			}
			// Skript mit Hilfe der de.bsvrz.sys.funclib.dynobj-Bibliothek erzeugen, mit der der Konfigurationsbereich des neuen Objekte parametriert werden kann.
			try {
				script = DynamischeObjekte.getInstanz(_dav).erzeugeObjekt(dynamicObjectType, name, pid);
			}
			catch(DynObjektException e1) {
				// Wenn die de.bsvrz.sys.funclib.dynobj Bibliothek kein Objekt erzeugen konnte, dann wird versucht, das Objekt im Defaultbereich anzulegen
				final String message =
						"Neues dynamisches Objekt für das zu erstellende Skript konnte nicht mit der Bibliothek de.bsvrz.sys.funclib.dynobj erzeugt "
						+ "werden, mit der der Konfigurationsbereich des neuen Objekts parametriert werden kann. Die Parametrierung einer Zuordnung von "
						+ "typ.puaSkript zum gewünschten Konfigurationsbereich ist erforderlich.";
				throw new RuntimeException(message, e1);
//				try {
//					final String defaultAreaPid = configAuth.getConfigurationData(model.getAttributeGroup("atg.konfigurationsVerantwortlicherEigenschaften")).getTextArray( "defaultBereich").getText(0);
//					final ConfigurationArea defaultArea = model.getConfigurationArea(defaultAreaPid);
//					script = defaultArea.createDynamicObject(dynamicObjectType, pid, name);
//				}
//				catch(Exception e2) {
//					debug.warning("Neues dynamisches Objekt für das zu erstellende Skript konnte nicht im Defaultbereich des Konfigurationsverantwortlichen "
//					              + "erzeugt werden.", e2);
//					throw new RuntimeException("Neues dynamisches Objekt für das zu erstellende Skript konnte weder mit der Bibliothek "
//					                           + "de.bsvrz.sys.funclib.dynobj noch im Defaultbereich der AOE erzeugt werden", e2);
//				}
			}
			MutableSet scriptSet = _configAuth.getMutableSet(PidScript.objectSet);
			scriptSet.add(script);
			this.recentlyCreatedScript = script;
		}
		catch(Exception e) {
			debug.warning("Fehler beim erzeugen eines dynamisches Objekt für ein neues Skript mit Pid '" + pid + "' und Name '" + name + "'", e);
			script = null;
		}
		return script;
	}

	/**
	 * Liefert den Quelltext eines Skripts zurück. Blockiert, bis die Antwort eintrifft. Sendet die Anfrage als einfacher Sender.
	 *
	 * @param script Skriptobjekt dessen Quelltext geliefert werden soll.
	 * @param source Buffer, in den der Skriptquelltext abgelegt wird.
	 * @param meta   Metainformationen über das Skript.
	 *
	 * @return Ergebnis der Operation.
	 */
	public ProtocolRequestResult getSource(DynamicObject script, StringBuffer source, AtlMeta meta) {
		ScriptAnswerReceiver receiver = null;
		try {
			long requestId = getRequestId();
			// schon vorher auf Antwort anmelden
			receiver = new ScriptAnswerReceiver(_dav, script, _client, requestId, _timeout);
			receiver.subscribe(ReceiverRole.receiver());

			AtlExtra extra = new AtlExtra(_client, requestId, ScriptRelatedOperation.SOURCE, "");
			Data requestData = AtgScriptRequest.build(_dav, new AtlBase(), extra);
			if(SimpleSender.sendWait(_dav, script, _ddScript, requestData, _timeout)) {
				//jetzt auf das Ergebnis warten & vergleichen.
				ResultData rd = receiver.take();
				if(rd == null) {
					return new ProtocolRequestResult(ErrorMessage.TIMEOUT);
				}
				else {
					AtgAnswer answer = AtgAnswer.getJavaObject(rd.getData());
					if(answer.atlExtra.opCode == ScriptRelatedOperation.ACK_FLAG) {
						source.append(answer.atlExtra.source);
						meta.set(answer.atlMeta);
						return new ProtocolRequestResult();
					}
					else {
						return new ProtocolRequestResult(answer.message);
					}
				}
			}
			else {
				return new ProtocolRequestResult(ErrorMessage.CAN_NOT_SEND);
			}
		}
		catch(ConfigurationException e) {
			return new ProtocolRequestResult(ErrorMessage.COMMUNICATION + e.getMessage());
		}
		catch(SenderException e) {
			return new ProtocolRequestResult(e.getMessage());
		}
		catch(InterruptedException e) {
			return new ProtocolRequestResult(ErrorMessage.INTERRUPTED_EXCEPTION + e.getMessage());
		}
		finally {
			if(receiver != null) receiver.unsubscribe();
		}
	}


	/**
	 * Liefert die Standardwerte und die definierbaren Aspekte/Aliase eines Skripts zurück. Blockiert, bis die Antwort eintrifft. Sendet die Anfrage als
	 * einfacher Sender.
	 *
	 * @param script Skriptobjekt dessen Standardwerte geliefert werden sollen.
	 * @param defaults Hier werden die Standardwerte abgelegt. Es ist ein leeres AtlDefaults-Objekt zu übergeben.
	 * @param meta   Hier werden die Metainformationen über das Skript abgelegt. Es ist ein leeres AtlMeta-Objekt zu übergeben.
	 *
	 * @return Ergebnis der Operation.
	 */
	public ProtocolRequestResult getDefaults(final DynamicObject script, final AtlDefaults defaults, final AtlMeta meta) {
		ScriptAnswerReceiver receiver = null;
		try {
			long requestId = getRequestId();
			// schon vorher auf Antwort anmelden
			receiver = new ScriptAnswerReceiver(_dav, script, _client, requestId, _timeout);
			receiver.subscribe(ReceiverRole.receiver());

			AtlExtra extra = new AtlExtra(_client, requestId, ScriptRelatedOperation.DEFAULTS, "");
			Data requestData = AtgScriptRequest.build(_dav, new AtlBase(), extra);
			if(!requestData.isDefined()){
				return new ProtocolRequestResult(ErrorMessage.OLD_DATAMODEL);
			}
			if(SimpleSender.sendWait(_dav, script, _ddScript, requestData, _timeout)) {
				//jetzt auf das Ergebnis warten & vergleichen.
				ResultData rd = receiver.take();
				if(rd == null) {
					return new ProtocolRequestResult(ErrorMessage.TIMEOUT);
				}
				else {
					AtgAnswer answer = AtgAnswer.getJavaObject(rd.getData());
					if(answer.atlExtra.opCode == ScriptRelatedOperation.ACK_FLAG) {
						defaults.set(answer.atlExtra.getDefaults());
						meta.set(answer.atlMeta);
						return new ProtocolRequestResult();
					}
					else {
						return new ProtocolRequestResult(answer.message);
					}
				}
			}
			else {
				return new ProtocolRequestResult(ErrorMessage.CAN_NOT_SEND);
			}
		}
		catch(ConfigurationException e) {
			return new ProtocolRequestResult(ErrorMessage.COMMUNICATION + e.getMessage());
		}
		catch(SenderException e) {
			return new ProtocolRequestResult(e.getMessage());
		}
		catch(InterruptedException e) {
			return new ProtocolRequestResult(ErrorMessage.INTERRUPTED_EXCEPTION + e.getMessage());
		}
		finally {
			if(receiver != null) receiver.unsubscribe();
		}
	}


	/**
	 * Entfernt ein Skriptobjekt und den zugehörigen Skriptquelltext. Löscht Skript aus der Menge der Skriptobjekte.
	 *
	 * @param script Skriptobjekt das gelöscht wird.
	 *
	 * @return Ergebnis der Anfrage
	 */
	public ProtocolRequestResult deleteScript(DynamicObject script) {
		if(script != null) {
			try {
				MutableSet scriptSet = _configAuth.getMutableSet(PidScript.objectSet);
				scriptSet.remove(script);
			}
			catch(Exception e) {
				return new ProtocolRequestResult(e);
			}
		}
		return new ProtocolRequestResult();
	}

	/**
	 * Informiert die Listener über die Fertigstellung eines Protokolls
	 * @param jobInProgess Auftrags-Information
	 */
	void protocolFinished(final JobInProgress jobInProgess) {
		synchronized(_protocolListeners){
			for(final ProtocolListenerEx listener : _protocolListeners) {
				final Runnable asyncProtocolFinishedNotifier = new Runnable() {
					@Override
					public void run() {
						listener.jobFinished(jobInProgess);
					}
				};
				new Thread(asyncProtocolFinishedNotifier).start();
			}
		}
	}

	/**
	 * Informiert die Listener über das Löschen eines Protokolls
	 * @param id Protokoll-ID
	 */
	void protocolRemoved(final long id) {
		synchronized(_protocolListeners){
			for(final ProtocolListenerEx listener : _protocolListeners) {
				final Runnable asyncProtocolFinishedNotifier = new Runnable() {
						@Override
						public void run() {
							listener.protocolRemoved(id);
						}
					};
					new Thread(asyncProtocolFinishedNotifier).start();
			}
		}
	}

	/**
	 * Informiert die Listener über das Gelesenwerden eines Protokolls
	 * @param id Protokoll-ID
	 */
	void protocolRead(final long id) {
		synchronized(_protocolListeners){
			for(final ProtocolListenerEx listener : _protocolListeners) {
				final Runnable asyncProtocolFinishedNotifier = new Runnable() {
						@Override
						public void run() {
							listener.protocolRead(id);
						}
					};
					new Thread(asyncProtocolFinishedNotifier).start();
			}
		}
	}

	/**
	 * Ruft eine liste mit aktuell in Bearbeitung befindlichen Aufträgen ab.
	 * @param resultInfo  Wird mit Ergebnis der Operation belegt. Darf null sein.
	 * @return Liste mit allen protokollbezogenen Aufträgen, die sich PuA-seitig in der Warteschlange oder in Bearbeitung befinden.
	 * Für jeden Auftral ist eine Auftrags-ID (JobId) abrufbar, mit der Auftrag z. B. mit {@link #abortProtocol(long)} abgebrochen werden kann.
	 */
	public List<JobInProgress> getJobList(ProtocolRequestResult resultInfo) {
		if(resultInfo == null) resultInfo = new ProtocolRequestResult();

		long requestId = getRequestId();
		ProtocolAnswerReceiver receiver = null;
		AtgProtocolRequest answer = null;
		//Auf Antwort anmelden
		try {
			receiver = new ProtocolAnswerReceiver(_dav, _client, requestId, false, _timeout);
			receiver.subscribe(ReceiverRole.drain());
			resultInfo.set(sendRequest(requestId, ProtocolRelatedOperation.JOB_LIST, new byte[0]));
			if(resultInfo.isRequestSuccessful()) {
				ResultData rd = receiver.take();
				if(rd == null) {
					resultInfo.set(ErrorMessage.TIMEOUT);
				}
				else {
					answer = AtgProtocolRequest.getJavaObject(rd.getData());
					if(answer.opCode != ProtocolRelatedOperation.ACK_FLAG) {
						resultInfo.set((String)SerializerUtil.deserialize(answer.requestData));
					}
					else {
						resultInfo.setSuccessful();
						return deserializeJobList(answer.requestData);
					}
				}
			}
		}
		catch(FailureException e) {
			resultInfo.set(e.getMessage());
		}
		catch(ConfigurationException e) {
			resultInfo.set(ErrorMessage.COMMUNICATION + e.getMessage());
		}
		catch(Exception e) {
			e.printStackTrace();
			resultInfo.set(ErrorMessage.UNEXPECTED_DATA_TYPE + ((answer != null) ? SerializerUtil.deserialize(answer.requestData).getClass().getName() : ""));
		}
		finally {
			if(receiver != null) receiver.unsubscribe();
		}
		return null;
	}

	private List<JobInProgress> deserializeJobList(final byte[] requestData) {
		List<JobInProgress> jobs = (List<JobInProgress>) SerializerUtil.deserialize(requestData);
		for(JobInProgress job : jobs) {
			// Datenmodell wurde nicht serialisiert und muss hier ergänzt werden
			job.initialize(_model);
		}
		return jobs;
	}

	/**
	 * Interne Methode, die von {@link de.bsvrz.pua.prot.client.ProtocolResultStream} aus aufgerufen wird um die
	 * automatische Flusskontrolle zu triggern
	 * @param jobId Auftrags-ID
	 * @param targetNumDatasets Nummer des Datensatzes bis zu dem der Server Daten versenden soll
	 */
	void sendFlowControl(final long jobId, final long targetNumDatasets) {
		Runnable task = new Runnable() {
			@Override
			public void run() {
				try {
					sendRequest(jobId, ProtocolRelatedOperation.FLOW_CONTROL, SerializerUtil.serializeId(targetNumDatasets));
				}
				catch(Exception e) {
					debug.fine("Exception beim Versenden der Flusskontrolle", e);
				}
			}
		};
		_flowControlExcecutor.submit(task); // Flusskontrolle asynchron versenden um den Emfang von Daten nicht zu bremsen
	}

	/**
	 * Gibt die Größe des "Sliding Window" für die automatische Flusskontrolle zurück
	 * @return die Größe des "Sliding Window" für die automatische Flusskontrolle, d.h. die maximale Anzahl der Datensätze, die
	 * gleichzeitig übertragen werden.
	 */
	public int getFlowControlWindowSize() {
		return _flowControlWindowSize;
	}

	/**
	 * Setzt die Größe des "Sliding Window" für die automatische Flusskontrolle. Der PuaClient verwendet beim Protokollabruf zur
	 * Flusskontrolle ein einfaches "Sliding Window"-Verfahren, dass die Anzahl der Protokolldatensätze, die gleichzeitig übertragen werden
	 * beschränkt. Hierzu teilt der Client dem PuA-Server regemäßig mit, wie viele Protokolldatensätze er bereits empfangen hat, plus wie
	 * viele der PuA-Server zusätzlich senden soll.
	 *
	 * @param flowControlWindowSize
	 * Maximale Anzahl Protokoll-Ergebnis-Datensätze, die gleichzeitig von PuA-Server bis zum PuaClient gesendet werden.
	 * Auf 0 setzen um Flusskontrolle abzuschalten.
	 *
	 */
	public void setFlowControlWindowSize(final int flowControlWindowSize) {
		if(flowControlWindowSize < 0) throw new IllegalArgumentException("flowControlWindowSize muss >= 0 sein");
		_flowControlWindowSize = flowControlWindowSize;
	}

	/**
	 * Gibt belegte Ressourcen frei und meldet Datenverteiler-Anmeldungen ab.
	 * Nach dem Aufruf dieser methode sollte diese Instanz nicht weiter verwendet werden. 
	 */
	@Override
	public void close() {
		if(_closed) return;
		try {
			startStopProtocolListener((byte) 0);
		}
		catch(FailureException e) {
			debug.warning("Fehler beim Abmelden des Protokoll-Listeners", e);
		}
		ConnectionManager.unsubscribeSender(_dav, _puaSendControlListener, _configAuth, _ddProtocol);
		_protocolAnswerReceiver.unsubscribe();
		_flowControlExcecutor.shutdown();
		_closed = true;
	}

	void updateArsStatus(final boolean status) {
		_archiveOnline = status;
		for(StatusListener statusListener : _statusListeners) {
			final Runnable runnable = new Runnable() {
				@Override
				public void run() {
					statusListener.statusChanged(_listenerOnline, _archiveOnline);
				}
			};
			new Thread(runnable).start();
		}
	}

	/**
	 * Setzt die Version des vom Servers unterstützten Übertragungsprotokolls.
	 * @param protocolVersion (Siehe Protokollversionen.md)
	 */
	void incrementProtocolVersion(final int protocolVersion) {
		_protocolVersion = Math.max(_protocolVersion, protocolVersion);
	}

	/**
	 * Gibt die vom Server aktuell unterstützte Protokollversion zurück. Die Protokollversion wird nach dem Erstellen des
	 * PuaClients ausgehandelt, sodass die präzise Version möglicherweise erst wenige Sekunden nach der Initialisierung dieser
	 * Klasse zur Verfügung steht.
	 * @return Aktuell vom Server unterstützte Protokollversion 
	 */
	public int getServerProtocolVersion() {
		return _protocolVersion;
	}

	private static class ProtocolListenerWrapper implements ProtocolListenerEx {
		private final ProtocolListener _listener;

		public ProtocolListenerWrapper(final ProtocolListener listener) {
			_listener = listener;
		}

		@Override
		public void jobFinished(final JobInProgress job) {
			_listener.protocolFinished(job.getJobId());
		}

		@Override
		public void protocolRemoved(final long protocolId) {
		}

		@Override
		public void protocolRead(final long protocolId) {
		}

		public ProtocolListener getListener() {
			return _listener;
		}

		@Override
		public boolean equals(final Object o) {
			if(this == o) return true;
			if(o == null || getClass() != o.getClass()) return false;

			final ProtocolListenerWrapper that = (ProtocolListenerWrapper) o;

			return _listener == that._listener;

		}

		@Override
		public int hashCode() {
			return _listener.hashCode();
		}
	}

	private class PuaSendControlListener implements ClientSenderInterface {

		@Override
		public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {
			try {
				synchronized(this) {
					if(state == START_SENDING) {
						// Übertragung der Versionsnummer
						startStopProtocolListener((byte) 3, (byte)CURRENT_PROTOCOL_VERSION);
						// Erweiterte Anmeldung für neuere PuA-Server-Versionen
						ProtocolRequestResult extendedResult = startStopProtocolListener((byte) 2);
						if(!extendedResult.isRequestSuccessful()) {
							debug.warning(
									"PuA-Server unterstützt erweiterte Benachrichtigungen nicht und sollte aktualisiert werden", extendedResult
											.getErrorMsg()
							);
						}
						else {
							// Server versteht Listener-Anmeldung -> Protokollversion ist mindestens 2
							incrementProtocolVersion(2);
						}
						ProtocolRequestResult result = startStopProtocolListener((byte) 1);
						_listenerOnline = result.isRequestSuccessful();
						if(_listenerOnline) {
							updateArsStatus(_archiveOnline);
						}
					}
					else {
						if(_listenerOnline) {
							_listenerOnline = false;
							updateArsStatus(false);
						}
					}
					_listenerRequested = true;
					this.notifyAll();
				}
			}
			catch(FailureException e) {
				debug.warning("Kann Listener nicht an PuA-Server anmelden", e);
			}
		}

		@Override
		public boolean isRequestSupported(final SystemObject object, final DataDescription dataDescription) {
			return true;
		}
	}

	@Override
	public String toString() {
		return "PuaClient Protokollversion: " + getServerProtocolVersion();
	}
}
