/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.interpreter;

import antlr.ANTLRException;
import antlr.CommonAST;
import antlr.RecognitionException;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.pua.prot.util.ProcessingInformation;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.pua.prot.util.ProtocolRequestResult;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;

import java.io.DataInputStream;

/**
 * Klasse zur Steuerung des Skript-Interpreter Moduls. Diese Klasse steuert die Abläufe beim Interpretieren des Skripts.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class Interpreter implements InterpreterInterface {

	/** Verbindung zum Datenverteiler */
	private DataModel _model;

	/**
	 * Erzeugen des Interpreters
	 *
	 * @param model Verbindung zur Konfiguration.
	 */
	public Interpreter(final DataModel model) {
		this._model = model;
	}

	/**
	 * Startet die Interpretation des Skripts.
	 *
	 * @param script    Quelltext Skript
	 * @param fileName  Dateiname des Skripts (nur für detailliertere Fehlerberichte nötig)
	 * @param parameter Informationen zum Auswerten des Skripts. Falls <code>null</code> wird das Skript nur intrepretiert und nicht ausgeführt.
	 * @param pi        Objekt, in dem die gewonnenen Informationen abgelegt werden
	 *
	 * @return Informationen über die Ausführung der Operation
	 */
	public ProtocolRequestResult process(final DataInputStream script, final String fileName, @Nullable final ProcessingParameter parameter, final ProcessingInformation pi) {

		try {
			// attach PuA lexer to the input stream,
			ProtocolLexer lexer = new ProtocolLexer(script);
			// Create parser attached to lexer
			ProtocolParser parser = new ProtocolParser(lexer);
			parser.setFilename(fileName);
			parser.init(_model, parameter, pi);
			// Start up the parser by calling the rule to start with
			parser.skript();
			// Get the tree out of the parser
			final CommonAST t = (CommonAST)parser.getAST();

			// Zum Anzeigen des erzeugten ASTs:
			//CommonAST.setVerboseStringConversion(true,parser.getTokenNames());
/*			antlr.ASTFactory factory= new antlr.ASTFactory();
			antlr.collections.AST r= factory.create(0,"AST ROOT");
			r.setFirstChild(t);
			final antlr.debug.misc.ASTFrame frame = new antlr.debug.misc.ASTFrame("AST von "+fileName, r);
			frame.setVisible(true);
			frame.addWindowListener(
				new java.awt.event.WindowAdapter() {
				   public void windowClosing (java.awt.event.WindowEvent e) {
					   frame.setVisible(false); // hide the Frame
					   frame.dispose();
					   System.exit(0);
				   }
				}
			);			
*/

			//Walk the tree
			final ProtocolTreeWalker walker = new ProtocolTreeWalker(parser.getSemantics());
			// Traverse the tree created by the parser
			walker.skript(t);
			final boolean process = parameter != null;
			parser.getSemantics().cleanUp(parameter, process);
			return new ProtocolRequestResult();
		}
		catch(RecognitionException e) {
			return errorMsg(
					e,
					InterpreterErrorMessage.ERROR_IN + fileName + ", " + InterpreterErrorMessage.ERROR_AT + e.line + InterpreterErrorMessage.DELIMITER
					+ e.getMessage()
			);
		}
		catch(ANTLRException e) {
			return errorMsg(e, InterpreterErrorMessage.ERROR_IN + fileName + ", " + InterpreterErrorMessage.DELIMITER + e.toString());
		}
	}

	/**
	 * Liefert den Rückgabewert.
	 *
	 * @param e   Auslösende Exception.
	 * @param msg Nachricht.
	 *
	 * @return Aus der Exception und Nachricht erzeugter Rückgabewert.
	 */
	private ProtocolRequestResult errorMsg(final Exception e, final String msg) {
		if(e.getMessage() != null) {
			return new ProtocolRequestResult(msg);
		}
		else {
			final StringBuffer sb = new StringBuffer();
			for(int i = 0; i < e.getStackTrace().length; i++) {
				sb.append(e.getStackTrace()[i].toString() + "\n");
			}
			return new ProtocolRequestResult(sb.toString());
		}
	}
}
