/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.interpreter.semantics;

import antlr.Token;
import antlr.collections.AST;
import de.bsvrz.pua.prot.exceptions.CriticalParserException;
import de.bsvrz.pua.prot.exceptions.SemanticErrorException;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.util.ProcessingInformation;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.pua.prot.util.attributes.DurationAttributeDescription;
import de.bsvrz.pua.prot.util.attributes.RealAttributeDescription;
import de.bsvrz.pua.prot.util.attributes.TempAttributeDescription;

import java.util.*;

/**
 * Überprüft die Semantik von Aggregationen.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class AggregationSemantics implements SemanticChecks {

	private Semantics _semantics;

	private AttributeSemantics _attributeSemantics;

	private RealAttributeSemantics _realAttributeSemantics;

	private SymbolTable _symbolTable;

	/** Zeigt ob Aggregationen im Skript verwendet werden. */
	private boolean _aggregationsUsed = false;

	/**
	 * Konstruktor
	 *
	 * @param semantics              Hauptobjekt der Semantiküberprüfung.
	 * @param symbolTable            Symboltabelle.
	 * @param attributeSemantics     Semantiküberprüfung für die Attribute.
	 * @param realAttributeSemantics Semantiküberprüfung für reale Attribute.
	 */
	public AggregationSemantics(
			final Semantics semantics, final SymbolTable symbolTable, final AttributeSemantics attributeSemantics, final RealAttributeSemantics realAttributeSemantics) {
		this._semantics = semantics;
		this._attributeSemantics = attributeSemantics;
		this._realAttributeSemantics = realAttributeSemantics;
		this._symbolTable = symbolTable;
	}

	/**
	 * Fügt einem Attribut eine Aggregation hinzu.
	 *
	 * @param ast                Knoten (wird nur für Zeilenangabe in der Fehlermeldung benötigt)
	 * @param attributeQualifier AttributeQualifier
	 * @param aggregations Aggregations
	 * @param padData            <code>true</code> falls Daten durch die Datenaufbereitung aufgefüllt werden sollen.
	 *
	 * @throws CriticalParserException Kommunikationsfehler oder falls die Qualifizierung nicht korrekt ist oder falls die angegebene Aggregationsfunktion nicht
	 *                                 existiert
	 */
	public void addAggregation(final AST ast, final List<String> attributeQualifier, final List<String> aggregations, final boolean padData)
			throws CriticalParserException {
		try {
			if(attributeQualifier == null || attributeQualifier.size() < 1) throw new SemanticErrorException(InterpreterErrorMessage.NO_SCOPE_OR_FUNCTION);


			for(final String functionName : aggregations) {
				_symbolTable.addAggregationDefinition(new Token(), functionName, functionName);

				// Überprüfen, ob die Aggregation überhaupt definiert wurde:
				if(!_symbolTable._aggregationFcts.containsKey(functionName)) {
					throw new SemanticErrorException(InterpreterErrorMessage.UNDEFINED_AGGREGATION + functionName);
				}
			}

			//es kann sich um das Zeitdauer Attribut handeln
			if(attributeQualifier.size() == 1 && attributeQualifier.get(0).equals(DurationAttributeDescription.DURATION)) {
				if(_symbolTable._durationAttribute == null)					//falls das zeitdauer-Attribut noch nicht definiert wurde.
				{
					throw new CriticalParserException(InterpreterErrorMessage.DURATION_ATTRIBUTE_NOT_DEFINED);
				}
				for(final String functionName : aggregations) {
					_symbolTable._durationAttribute.addAggregation(_symbolTable._aggregationFcts.get(functionName));
				}
			}
			else if(attributeQualifier.size() == 1
			        && _symbolTable.isTempAttributeDefined(attributeQualifier.get(0))) {	//es kann sich um ein temporäres Attribut handeln...
				//Attribut aus Symboltabelle holen - da wurde es ja schon angelegt
				final TempAttributeDescription attribute = _symbolTable.getTempAttributeDefinition(attributeQualifier.get(0));
				// Dann Aggregation dort eintragen
				for(final String functionName : aggregations) {
					attribute.addAggregation(_symbolTable._aggregationFcts.get(functionName));
				}
			}
			else	//es handelt sich um ein reales Attribut
			{
				//Überprüfe ob Qualifizierung korrekt ist:
				final RealAttributeDescription[] attribute = _realAttributeSemantics.createRealAttributeDescription(attributeQualifier, padData);
				//Aggregationsinformationen in Attributbeschreibung eintragen
				for(final RealAttributeDescription rad : attribute) {
					for(final String functionName : aggregations) {
						rad.addAggregation(_symbolTable._aggregationFcts.get(functionName));
					}
				}

				//Aggregationsinformationen für späteren Gebrauch ablegen.
				_attributeSemantics.storeAggregationInformation(attribute);
			}

			//vermerken, dass Aggregationen im Skript verwendet werden.
			_aggregationsUsed = true;
		}
		catch(CriticalParserException criticalException) {
			criticalException.line = ast.getLine();
			throw criticalException;
		}
	}

	/**
	 * Fügt dem Zeitdauer Attribut eine Aggregation  hinzu.
	 *
	 * @param ast             Knoten (wird nur für die Zeilenangabe in der Fehlermeldung benötigt)
	 * @param aggregationName Einelementige Liste mit dem Namen der Aggregation
	 *
	 * @throws CriticalParserException Kommunikationsfehler
	 */
	public void addAggregationForDuration(final AST ast, final List<String> aggregationName) throws CriticalParserException {
		addAggregation(ast, Arrays.asList(DurationAttributeDescription.DURATION), aggregationName, false);
	}

	/**
	 * Legt die Anwendungsart der Aggregationen fest
	 *
	 * @param order Order
	 *
	 * @throws SemanticErrorException Falls Wiederholungen in der anwendung auftreten (etwa weil gleichzeitig sowohl gesamt(objekt) als auch gesamt(spalte)
	 *                                verwendet wurde)
	 */
	public void setOrder(final HashSet<ProcessingInformation.ApplyAggregation> order) throws SemanticErrorException {
		_semantics.pi.setOrder(order);
	}

	/**
	 * Überprüft, ob Aggregationen verwendet werden, falls das Schlüsselwort 'anwendung' im Skript verwendet wird.
	 *
	 * @see de.bsvrz.pua.prot.interpreter.semantics.SemanticChecks#cleanUp(de.bsvrz.pua.prot.util.ProcessingParameter,boolean)
	 */
	public void cleanUp(final ProcessingParameter pp, final boolean processScript) throws SemanticErrorException {
		if(!_semantics.pi.getOrder().isEmpty() && !_aggregationsUsed) throw new SemanticErrorException(InterpreterErrorMessage.ORDER_WO_AGGREGATION);
	}
}
