/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.manager.taskmanager;

import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.pua.prot.manager.ProtocolAnswer;
import de.bsvrz.pua.prot.manager.ProtocolRelatedOperation;
import de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface;
import de.bsvrz.pua.prot.manager.status.StatePublisher;
import de.bsvrz.pua.prot.processing.ProcessingInterface;
import de.bsvrz.pua.prot.sender.PuaSender;
import de.bsvrz.pua.prot.util.JobInProgress;
import de.bsvrz.pua.prot.util.JobState;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.PidProtocol;
import de.bsvrz.sys.funclib.losb.datk.SerializerUtil;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

/**
 * Thread der einen Auftrag der in den Bereich der Protokollerstellung fällt ausführt. Mögliche Aufträge sind:<br> - Protokoll erstellen<br> - Protokoll
 * abbrechen<br> - Status der Protokollerstellung anfordern<br>
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ThreadElement extends Thread {

	/** Status der Protokollerstellung nicht ermittelbar. */
	private static final byte INVALID_STATUS = -1;
	
	private AbstractTask _task;
	private PuaSender _sender;

	public SerializerUtil getSerializer() {
		return _serializer;
	}

	public ThreadPool getCreator() {
		return _creator;
	}

	public long getMaxDelay() {
		return _maxDelay;
	}

	public int getTimeout() {
		return _timeout;
	}

	/**
	 * Zeigt an, ob der Thread bereits läuft, d.h. ob er sich bereits in die {@code run} Methode befindet.
	 *
	 * @return Thread läuft?
	 */
	public boolean isRunning() {
		return running;
	}

	public void setRunning(boolean running) {
		this.running = running;
	}

	public enum Operation {

		NONE(ProtocolRelatedOperation.INVALID),
		CREATE_PROTOCOL(ProtocolRelatedOperation.CREATE),
		SEND_SAVED_PROTOCOL(ProtocolRelatedOperation.SEND_SAVED_PROTOCOL),
		SEND_UNREAD_PROTOCOL(ProtocolRelatedOperation.SEND_UNREAD_PROTOCOL);

		private final int _code;

		Operation(final int code) {
			_code = code;
		}

		public int getCode() {
			return _code;
		}

	}
	
	/** Objekt das die Datenaufbereitung durchführt */
	private ProcessingInterface _processor = null;

	/** Erzeuger diser ThreadElement Instanz */
	private ThreadPool _creator = null;

	/** True falls der Thread die Datenaufbereitung beenden soll. */
	private boolean _abort = false;

	/** Falls der Thread vollständig beendet werden soll. */
	private volatile boolean _die = false;

	/** Debug-Ausgabe */
	private static final Debug _debug = Debug.getLogger();

	private boolean running = false;

	/** Data Manager. Ermöglicht das Interpretieren des Skripts */
	private final DataManagerInterface _dataManager;


	/** Maximale Zeitdauer in ms, für die die Protokollerstellung angehalten werden kann. */
	private final long _maxDelay;

	/**
	 * Anzahl Minuten ohne Antwort vom Client, nach der die Übertragung getrennt wird
	 */
	private final int _timeout;

	/** true falls ein Auftrag vorliegt und noch abgearbeitet werden muss. */
	private boolean _busy;

	/** Objekt zum serialisieren und zippen der Ergebnisdaten. */
	private SerializerUtil _serializer = null;

	/**
	 * Initialisierung des Objekts
	 *
	 * @param creator     Erzeuger des Threads. Wird benachrichtigt (@link ThreadPool#finished(ThreadElement)}), sobald Thread unbeschäftigt ist.
	 * @param dataManager Data Manager. Ermöglicht das Interpretieren des Skripts
	 * @param maxDelay    Maximale Zeitdauer in ms, für die die Protokollerstellung angehalten werden kann.
	 *
	 * @param timeout Timeout
	 * @throws FailureException Fehler bei der Initialisierung.
	 */
	public ThreadElement(ThreadPool creator, DataManagerInterface dataManager, long maxDelay, final int timeout) throws FailureException {
		_timeout = timeout;
		_serializer = new SerializerUtil();
		this._creator = creator;
		this._dataManager = dataManager;
		this._maxDelay = maxDelay;
	}

	/**
	 * Liefert den Status des Auftrags
	 *
	 * @return Prozentualer Fertigstellungsgrad des Auftrags oder {@link #INVALID_STATUS} falls kein Auftrag anliegt.
	 */
	public byte getStatus() {
		if(_processor != null) {
			return _processor.getStatus();
		}
		else {
			return INVALID_STATUS;
		}
	}



	/**
	 * Bricht die Protokollerstellung ab. Der Thread meldet dem {@link #_creator}, dass er wieder für neue Anfragen zur Verfügung steht. Dazu ruft er die Methode
	 * {@link ThreadPool#finished(ThreadElement)} auf. Wird das Protokoll gespeichert auf dem Datenträger gespeichert, wird es wieder gelöscht. Im Gegensatz zu
	 * {@link #kill} steht der Thread danach weiterhin zur Verfügung.
	 */
	public void abort() {
		_debug.finer("Thread " + this.getName() + ": Protokollerstellung wird abgebrochen");
		_abort = true;
		if(_processor != null) _processor.abort();

		/*
				 * Wichtig: Hier nicht creator.finished() aufrufen!
				 * Dies wird NUR in der run() Methode getan,
				 * sonst kann das Objekt nicht mehr gelockt werden! (für synchronized)
				 */
	}

	public void stopProtocolTransmission() {
		if(_processor != null) _processor.stopProtocolTransmission();
	}


	protected PuaSender getPuaSender(AbstractTask task) {
		return new PuaSender(task.getDav(), task.getClient(), new DataDescription(task.getDav().getDataModel().getAttributeGroup(PidProtocol.atgProtocolRequest),
		                                                                          task.getDav().getDataModel().getAspect(PidProtocol.aspAnswer)),
		                     () -> {});
	}
	
	/** Startet den Thread und versetzt ihn sofort in einen Schlafzustand, aus dem er erst wieder geweckt wird, wenn eine Anfrage eingeht. */
	@Override
    public void run() {
		try {
			while(!_die) {
				try {
					_debug.finer("Thread " + this.getName() + " beendet Anfragenbearbeitung & wartet");
					running = true;
					//falls Thread nach Fertigstellung eines Protokolls sofort wieder
					//einen Auftrag zugewiesen wird ist es möglich, dass activate(...)
					//VOR erreichen von wait(...)aufgerufen wird -> es muss geprüft
					//werden, ob überhaupt gewartet werden soll.
					synchronized(this){
						while(!_busy)
						{
							this.wait();
						}
					}
					if(!_die && _busy) {
						_debug.finer("Thread " + this.getName() + " beginnt mit Anfragenbearbeitung");
						try {
							_sender = getPuaSender(_task);
							_processor = _task.process(_dataManager, this, _sender);
							if(_processor != null) {
								_processor.process(_abort);
							}
						}
						catch(FailureException fe) {
							sendError(
									new FailureException(
											"Fehler bei der Protokollerstellung durch Thread " + this.getName() + ", " + fe.getMessage(), fe.getLogLevel()
									)
							);
						}
					}
				}
				catch(InterruptedException e) {    //exception wird nicht gesetzt, da es sich nicht um einen Fehler handelt (InterruptedException wird geworfen, wenn die DA abgebrochen werden soll).
					_debug.finest(ErrorMessage.INTERRUPTED_EXCEPTION + _task);
				}
				catch(Throwable e){
					_debug.error("Unbehandelter Fehler bei der Bearbeitung einer Anfrage", e);
					sendError(
							new FailureException(
									"Fehler bei der Protokollerstellung durch Thread " + this.getName() + ", " + e.getMessage(), LoggerException.ERROR
							)
					);
				}
				interrupted(); // Interrupted-Status löschen

				if(_busy) {
					assert _sender.getDav() != null;
					StatePublisher.getInstance(_sender.getDav()).notifyThreadFinished(this);
				}

				//Abbrechen nicht nötig
				_abort = false;
				_processor = null;
				//Erledigt melden & sender abmelden
				if(_sender != null) {
					_sender = null;
				}
				this.setName("IdleThread (was " + this.getName() + ")");
				_busy = false;                //Wichtig! finshed ruft u.U. wieder activate... auf. Dort wird busy wieder auf true gesetzt! Deshalb MUSS buys=false VOR creator.finished stehen!
				_creator.finished(this);
			}
			_debug.finer("Thread " + this.getName() + " hat die Anfragenbearbeitung beendet.");

		}
		finally {
			if(_sender != null) {
				_sender = null;
			}
			this.setName("IdleThread (was " + this.getName() + ")");
			_serializer.done();
			_serializer = null;
		}
	}

	/**
	 * Sendet eine Fehlermeldung an den Auftraggeber und gibt eine Debug-Nachricht aus.
	 *
	 * @param exception Fehlermeldung.
	 */
	public void sendError(FailureException exception) {
		exception.log();
		try {
			_debug.finest("Setze Fehlermeldung ab: " + _task.getRequestId());
			ProtocolAnswer.sendError(_sender, _task.getRequestId(), exception.getMessage());
		}
		catch(FailureException e) {
			_debug.error(ErrorMessage.CAN_NOT_SEND + e.getMessage());
		}
	}



	public void activateTask(final AbstractTask task) throws FailureException {
		// Thread-Steuerung:
		this._busy = true;
		//Auftragsparameter übernehmen.
		this.setName(task.getClass().getSimpleName() + " Pid:" + task.getProtocolId());
		this._task = task;
		synchronized(this) {
			notifyAll();
		}
	}

	/** Bricht den Auftrag ab und beendet den Thread vollständig. */
	public void kill() {
		_die = true;
		abort();
		synchronized(this) {
			notifyAll();
		}
	}

	/**
	 * Liefert die Protokoll-Id.
	 *
	 * @return Id des Protokolls, das erstellt wird.
	 */
	public long getProtocolId() {
		return _task.getProtocolId();
	}

	/**
	 * Gibt die eindeutige ID dieses Auftrags zurück. Diese Id wird statt der Protokoll-ID verwendet, da je Protokoll u.U. mehrere
	 * Aufträge gleichzeitig ausgeführt werden können.
	 * @return ID des Auftrags.
	 */
	public long getJobId() {
		return _task.getJobId();
	}

	/** Flußsteuerung: Versand des Protokolls anhalten. */
	public void pauseProcessing() {
		if(_processor != null) _processor.pauseProcessing();
	}

	/** Flußsteuerung: Versand des Protokolls fortsetzen. */
	public void resumeProcessing() {
		if(_processor != null) _processor.resumeProcessing();
	}


	public void flowControl(final long numData) {
		if(_processor != null) _processor.flowControl(numData);
	}


	/** @return Liefert die initiale Anfrage-Id. */
	public long getRequestId() {
		return _task.getRequestId();
	}

	/**
	 * Gibt die aktuell durchgeführte Aktion zurück
	 * @return 1: Protokollerstellung. 2: Gespeichertes Protokoll senden. 3: Noch nicht abgerufenes Protokoll senden.*/
	public Operation getOperation() {
		if(_task == null) return Operation.NONE;
		return _task.getOperation();
	}

	/**
	 * Erstellt ein JobInProgress-Objekt zur Übertragung des Auftragsstatus an den Client
	 *
	 * @return JobInprogress
	 * @throws de.bsvrz.sys.funclib.losb.exceptions.FailureException Exception
	 */
	public JobInProgress toJobInProgress() throws FailureException {
		JobState state = JobState.InProgress;
		if(_processor != null){
			state = _processor.getJobState();
		}
		PuaSender sender = _sender;
		if(sender == null){
			throw new FailureException("Client-ID nicht ermittelbar", LoggerException.ERROR);
		}
		return new JobInProgress(_task.getJobId(), _task.getOperation().getCode(), state, getStatus(), _task.getStartTime(), sender.getObject().getId(), _task instanceof CreateProtocolTask ? ((CreateProtocolTask) _task).getPp() : _task.getProtocolId());
	}

}
