/*
 * Copyright 2014 by Kappich Systemberatung Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.processing;

import de.bsvrz.dav.daf.main.config.ClientApplication;
import de.bsvrz.dav.daf.main.config.DynamicObject;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.client.dataobject.ProtocolResult;
import de.bsvrz.pua.prot.client.dataobject.ProtocolResultData;
import de.bsvrz.pua.prot.exceptions.NoncriticalException;
import de.bsvrz.pua.prot.manager.ProtocolAnswer;
import de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface;
import de.bsvrz.pua.prot.manager.datamanager.ProtocolInputStream;
import de.bsvrz.pua.prot.sender.PuaSender;
import de.bsvrz.pua.prot.util.JobState;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.SerializerUtil;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;
import de.bsvrz.sys.funclib.losb.util.Util;

/**
 * Lies ein gespeichertes Protokoll ein und versendet es an einen Empfänger
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ProtocolReader extends BasicFlowControl implements ProcessingInterface {

	private static final Debug _debug = Debug.getLogger();

	/** Protokoll-Id */
	private final long _protocolId;

	/** Objekt zum Versand der Daten */
	private final PuaSender _sender;

	/** Zeigt ob der Vorgang abgebrochen wurde. */
	private volatile boolean _abort = false;

	/** Status der Datenaufbereitung */
	private byte _status = 0;

	private final SerializerUtil _serializer;

	/** Der Thread der den Vorgang ausführt. Wird für interrupt() benutzt um den Vorgang abzubrechen. */
	private final Thread _thread;

	private final DataManagerInterface _dataManager;
	
	/** Auftrags-ID */
	private final long _jobId;

	private volatile JobState _jobState = JobState.InProgress;

	/**
	 * Erzeugt ein Objekt zur Datenaufbereitung.
	 * @param sender      Objekt zum Versand der Daten.
	 * @param serializer Serializer
	 * @param protocolId  Id des zu erstellenden Protokolls.
	 * @param maxDelay    Maximale Zeitdauer in ms für die die Protokollerstellung angehalten werden kann.
	 * @param timeout Timeout
	 * @param jobId JobId
	 * @param initialFlowControl InitialFlowControl
	 * @param dataManager DataManager
	 */
	public ProtocolReader(
			final PuaSender sender,
			final SerializerUtil serializer,
			final long protocolId,
			final long maxDelay,
			final int timeout,
			final DataManagerInterface dataManager,
			final long jobId,
			final long initialFlowControl) {
		super(maxDelay, timeout, initialFlowControl);
		_serializer = serializer;
		this._sender = sender;
		this._protocolId = protocolId;
		_dataManager = dataManager;
		_jobId = jobId;
		_thread = Thread.currentThread();
	}

	/** @see ProcessingInterface#process(boolean) */
	public void process(final boolean abortProcessing) {
		this._abort = abortProcessing;
		ProtocolInputStream protocolInputStream = null;
		try {
			protocolInputStream = _dataManager.getUnreadProtocol(_protocolId);
			byte[] processingParameters = protocolInputStream.getProcessingParameters();
			ProcessingParameter pp = ProcessingParameter.deserialize(_sender.getDav().getDataModel(), processingParameters);

			SystemObject client = _sender.getObject();
			DynamicObject requestUser = null;
			if(client instanceof ClientApplication) {
				ClientApplication clientApplication = (ClientApplication) client;
				requestUser = clientApplication.getUser();
			}

			if(!pp.isPublic() && pp.getCreatorUser() != null && !pp.getCreatorUser().equals(requestUser)) {
				throw new FailureException("Von " + pp.getCreatorUser() + " erstelltes Protokoll " + _protocolId + " ist nicht öffentlich und darf daher von " + requestUser + " nicht abgerufen werden." , LoggerException.ERROR);
			}

			ProtocolResult pr = protocolInputStream.take();
			while(!_abort && pr != null) {
				flowControl();
				if(pr instanceof ProtocolResultData) {
					ProtocolResultData data = (ProtocolResultData) pr;
					_status = data.status;
				}
				if(!ProtocolAnswer.sendProtocol(_sender, _jobId, _serializer, pr)) {
					_debug.info("Protokollabruf abgebrochen", _jobId);
					_jobState = JobState.Failed;
					return;
				}
				datasetSend();
				pr = protocolInputStream.take();
			}
			//Endedatensatz schreiben.
			if(!_abort) {
				_jobState = JobState.Finished;
				ProtocolAnswer.sendAck(_sender, _jobId);
				if(!_sender.flush()) {
					_jobState = JobState.Failed;
				}
			}
			else {
				_jobState = JobState.Aborted;
				ProtocolAnswer.sendAbort(_sender, _jobId, ProcessingErrorMessage.ABORT_PROTOCOL);
			}
		}
		catch(InterruptedException ignored) {
			_jobState = JobState.Aborted;
			_debug.finest(ErrorMessage.INTERRUPTED_EXCEPTION + _protocolId);
			_abort = true;
		}
		catch(FailureException e){
			_jobState = JobState.Failed;
			_debug.warning("Fehler beim Abruf eines Protokolls", e);
			try {
				ProtocolAnswer.sendError(_sender, _jobId, e.getMessage());
			}
			catch(FailureException ignored) {
			}
		}
		catch(Exception e) {
			_jobState = JobState.Failed;
			_debug.warning("Unerwarteter Fehler beim Abruf eines Protokolls", e);
			try {
				ProtocolAnswer.sendError(_sender, _jobId, "Fehler beim Abruf eines Protokolls: " + Util.getStackTrace(e));
			}
			catch(FailureException ignored) {
			}
		}
		finally {
			if(protocolInputStream != null){
				if(_jobState == JobState.Finished){
					try {
						protocolInputStream.closeAndRename();
					}
					catch(NoncriticalException e) {
						_debug.warning("Protokoll kann nicht als gelesen markiert werden", e);
					}
				}
				else {
					protocolInputStream.close();
				}
			}
		}
	}

	/**
	 * Bricht die Datenaufbereitung ab.
	 *
	 * @see ProcessingInterface#abort()
	 */
	public void abort() {
		synchronized(this) {
			_abort = true;
			notifyAll();
		}

		// NotifyAll reicht nicht immer, da während Archivanfragen auf andere Objekte synchronisiert wird.
		// _archiveBuffer.abort() und _onlineBuffer.abort() rufen interrupt() auf ihre eigenen Threads auf, aber
		// nicht auf den ThreadElement-Thread, der ebenfalls hängen bleiben kann (in ArchiveUser.execute())
		_thread.interrupt();
	}

	/** @see ProcessingInterface#getStatus() */
	public byte getStatus() {
		return _status;
	}

	/** @see ProcessingInterface#getProtocolId() */
	public long getProtocolId() {
		return _protocolId;
	}

	@Override
	public long getJobId() {
		return _jobId;
	}

	@Override
	public JobState getJobState() {
		return _jobState;
	}

	@Override
	public void stopProtocolTransmission() {
		abort(); // Es wird nichts gespeichert, also identisch zu abort()
	}

	@Override
	protected Object getClient() {
		return _sender.getObject();
	}
}
