/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.processing.onlinebuffer;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.Dataset;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.pua.prot.util.attributes.RealAttributeDescription;

import java.util.Objects;

/**
 * Klasse erlaubt die Modifikation von Ergebnisdaten. Erlaubt das Setzen des Datenindex und der Daten. Ersteres ist bei {@link de.bsvrz.dav.daf.main.Dataset}
 * und den davon abgeleiteten Klassen (Archivdaten, OnlineDaten) nicht möglich.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ExtResultData {

	private long _dataIndex;

	private Dataset _dataSet;

	private Data _data;

	private long _archiveTime;

	private long _dataTime;


	/**
	 * Erstellt Ergebnisdatensatz mit leerem Archivzeitstempel.
	 *
	 * @param dataSet   Vollständiger Ergebnisdatensatz
	 * @param data      Relevante Daten.
	 * @param dataIndex Datindex.
	 */
	public ExtResultData(Dataset dataSet, Data data, long dataIndex) {
		this._data = data;
		this._dataIndex = dataIndex;
		this._dataSet = dataSet;
		this._archiveTime = -1;
		this._dataTime = dataSet.getDataTime();
	}

	/**
	 * Erstellt Ergebnisdatensatz.
	 *
	 * @param dataSet     Vollständiger Ergebnisdatensatz
	 * @param data        Relevante Daten.
	 * @param dataIndex   Datindex.
	 * @param archiveTime Archivzeitstempel.
	 */
	public ExtResultData(Dataset dataSet, Data data, long dataIndex, long archiveTime) {
		this(dataSet, data, dataIndex);
		this._archiveTime = archiveTime;
	}

	/** @return Liefert den Datenindex. */
	public long getDataIndex() {
		return _dataIndex;
	}

	/** @return Returns the data. */
	public Data getData() {
		return _data;
	}

	/** @return Returns the dataSet. */
	public Dataset getDataSet() {
		return _dataSet;
	}

	/**
	 * Sorgt dafür, dass nur das Unterdatum betrachtet wird, dass in der RealAttributeDescription angegeben ist
	 * @param r RealAttributeDescription
	 */
	public void selectData(final RealAttributeDescription r) {
		if(_dataSet != null && _dataSet.getData() != null) {
			_data = r.getAttributeIdentifier().getDataElement(_dataSet.getData());
		}
	}

	/**
	 * Liefert den Datenzeitstempel.
	 *
	 * @return Datenzeitstempel.
	 */
	public long getDataTime() {
		return _dataTime;
	}

	/**
	 * Liefert den Datentyp.
	 *
	 * @return Datentyp.
	 */
	public DataState getDataType() {
		return _dataSet.getDataType();
	}

	/**
	 * Liefert den Archivdatentyp.
	 *
	 * @return Archivdatentyp.
	 */
	public ArchiveDataKind getDataKind() {
		return _dataSet.getDataKind();
	}

	/** @see java.lang.Object#equals(java.lang.Object) */
	@Override
    public boolean equals(Object obj) {
		if(obj instanceof ExtResultData) {
			ExtResultData other = (ExtResultData)obj;
			return Objects.equals(_dataSet, other._dataSet) && Objects.equals(_data, other._data) && _dataIndex == other._dataIndex;
		}
		else {
			return false;
		}
	}

	/** @return Liefert die Archivzeit zurück. Hat bei Online-Daten den Wert <code>-1</code>sein. */
	public long getArchiveTime() {
		return _archiveTime;
	}

	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		return _dataSet + " -> " + _data;
	}

	/**
	 * Setzt den Zeitstempel
	 *
	 * @param dts Zeitstempel
	 */
	public void setDataTime(long dts) {
		this._dataTime = dts;
	}

	/**
	 * Setzt den Datenindex
	 *
	 * @param di Datenindex
	 */
	public void setDi(long di) {
		_dataIndex = di;
	}
}
