/*
 * Copyright 2018 by Kappich Systemberatung Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.util;

import java.time.Duration;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;

/**
 * Diese Klasse repräsentiert einen zusammenhängenden Zeitbereich
 *
 * @author Kappich Systemberatung
 */
public final class DiscretePeriod {

	/** von */
	private final Instant _from;
	
	/** bis */
	private final Instant _to;

	/** 
	 * Erstellt eine neue DiscretePeriod zwischen zwei Zeitpunkten ({@link Instant}-Objekten)
	 * @param from von
	 * @param to  bis
	 */
	private DiscretePeriod(final Instant from, final Instant to) {
		_from = from;
		_to = to;
	}

	/**
	 * Erstellt eine DiscretePeriod basierend auf Millisekunden-Zeitstempeln seit Epoch
	 * @param fromMillis von
	 * @param toMillis   bis
	 * @return DiscretePeriod
	 */
	public static DiscretePeriod of(long fromMillis, long toMillis) {
		return of(Instant.ofEpochMilli(fromMillis), Instant.ofEpochMilli(toMillis));
	}

	/**
	 * Erstellt eine DiscretePeriod basierend auf {@link LocalDateTime}-Objekten in der Standard-Zeitzone.
	 * @param from von
	 * @param to   bis
	 * @return DiscretePeriod
	 */
	public static DiscretePeriod of(LocalDateTime from, LocalDateTime to) {
		return of(from, to, ZoneId.systemDefault());
	}

	/**
	 * Erstellt eine DiscretePeriod basierend auf {@link LocalDateTime}-Objekten in der angegebenen Zeitzone.
	 * @param from von
	 * @param to   bis
	 * @param zone Zeitzone
	 * @return DiscretePeriod
	 */
	public static DiscretePeriod of(LocalDateTime from, LocalDateTime to, final ZoneId zone) {
		return new DiscretePeriod(from.atZone(zone).toInstant(), to.atZone(zone).withLaterOffsetAtOverlap().toInstant());
	}

	/**
	 * Erstellt eine DiscretePeriod basierend auf {@link Instant}-Objekten.
	 * @param from von
	 * @param to   bis
	 * @return DiscretePeriod
	 */
	public static DiscretePeriod of(Instant from, Instant to) {
		if(to.isBefore(from)) {
			throw new IllegalArgumentException("to < from: " + to + " < " + from);
		}
		return new DiscretePeriod(from, to);
	}

	/**
	 * Gibt den Start-Zeitpunkt zurück
	 * @return Start
	 */
	public Instant getFrom() {
		return _from;
	}


	/**
	 * Gibt den End-Zeitpunkt zurück
	 * @return Bis-Zeit
	 */
	public Instant getTo() {
		return _to;
	}

	/**
	 * Gibt die Dauer des Bereichs zurück
	 * @return Dauer
	 */
	public Duration getDuration() {
		return Duration.between(_from, _to);
	}

	@Override
	public boolean equals(final Object o) {
		if(this == o) return true;
		if(o == null || getClass() != o.getClass()) return false;

		final DiscretePeriod that = (DiscretePeriod) o;

		if(!_from.equals(that._from)) return false;
		return _to.equals(that._to);
	}

	@Override
	public int hashCode() {
		int result = _from.hashCode();
		result = 31 * result + _to.hashCode();
		return result;
	}

	@Override
	public String toString() {
		return _from + " – " + _to;
	}
}
